--
--	Wholly
--	Written by scott@mithrandir.com
--
--	This was inspired by my need to replace EveryQuest with something that was more accurate.  The pins
--	was totally inspired by QuestHubber which I found when looking for a replacement for EveryQuest.  I
--	initially made a port of QuestHubber to QuestHubberGrail to make use of the Grail database, but now
--	I have integrated that work into this addon.  Many thanks for all the work put into QuestHubber.  I
--	was inspired to add a quest breadcrumb area from seeing one in Quest Completist.
--
--	Version History
--		001	Initial version.
--		002	Support for displaysDungeonQuests now works properly.
--			Added the ability for the panel tooltip to display questgivers.
--			Added the ability to click a quest in the panel to create a TomTom waypoint.
--			Map pins are only displayed for the proper dungeon level of the map.
--		003	Added a panel button to switch to the current zone.
--			Changed the Close button into a Sort button that switches between three different modes:
--				1. Alphabetical  2. Quest level, then alphabetical  3. Quest status, then alphabetical
--			Made map pins only have one pin per NPC, indicating the "best" color possible.
--			The entire zone dropdown button on the quest log panel now can be clicked to change zones.
--			Corrected a problem where someone running without LibStub would get a LUA error.
--			Corrected a localization issue.
--		004	Added the ability to display a basic quest prerequisite chain in the quest panel tooltip, requiring Grail 014 or later.
--			Added the ability to right-click a "prerequisite" quest in the panel to put a TomTom arrow to the first prerequisite quest.
--			Added Dungeons and Other menu items for map areas in the quest log panel.
--			The last-used sort preference is stored on a per-character basis.
--		005	Corrected the fix introduced in version 003 putting the LibDataBroker icon back in place.
--			Corrected a problem where the quest log tooltip would have an error if the questgiver was in a dungeon appearing in the zone.
--			Added the ability for the quest log tooltip to show that the questgiver should be killed (like the map pin tooltip).
--			The problem where map pins would not live update unless the quest log panel was opened has been fixed as long
--				as the Wholly check button appears on the map.
--			Added the ability to show quest breadcrumb information when the Quest Frame opens, showing a tooltip of breadcrumb
--				quests when the mouse enters the "breadcrumb area", and putting TomTom waypoints when clicking in it.
--		006	Added the new quest group "World Events" which has holiday quests in it, requiring Grail 015.
--			Added a tooltip to the Wholly check button on the map that indicates how many quests of each type are in the map.
--			Added a tooltip to the LibDataBroker icon that shows the quest log panel "map" selection and the quest count/type.
--			Added a tooltip to the quest log panel Zone button that shows the quest count/type.
--			Corrected the problem where the quest log panel and map pins were not live updating when quest givers inside dungeons checked.
--			Corrected the problem where an NPC that drops items that starts more than one quest does not display the information properly
--				in its tooltip.
--			Made it so the open World Map can be updated when crossing into a new zone.
--		007	Added the ability to show whether quests in the quest log are completed or failed.
--			Made it so right-clicking an "in log" quest will put in TomTom waypoints for the turn in locations, which requires Grail 016
--				for proper functioning since Grail 015 had a bug.
--			Made the strings for the preferences color quest information like it appears in the UI.
--			Made it so alt-clicking a log in the Wholly quest log selects the NPC that gives the quest or for the case of "in log" quests
--				the one to which the quest is turned in.
--		008	Split out Dungeons into dungeons in different continents, requiring Grail version 017.
--			Corrected a misspelling of the global game tooltip name.
--		009	Added localization for ptBR in anticipation of the Brazilian release.
--			Changed over to using Blizzard-defined strings for as many things as possible.
--			Corrected a problem that was causing the tooltip for creatures that needed to be killed to start a quest not to appear properly.
--			Added a few basic localizations.
--			Made the breadcrumb frame hide by default to attempt to eliminate an edge case.
--			Fixed a problem where button clicking behavior would never be set if the button was first entered while in combat.
--			Made prerequisite information appear as question marks instead of causing a LUA error in case the Grail data is lacking.
--		010	Made it so the color of the breadcrumb quest names match their current status.
--			The click areas to the right and bottom of the quest log window no longer extend past the window.
--			Added menu options for Class quests, Profession quests, Reputation quests, and Daily quests.  The Class and Profession quests will show all the quests in the system except for the class and professions that match the player.  For those, the quests are displayed using the normal filtering rules.  The Reputation quests follow the normal filtering rules except those that fail to be acceptable solely because of reputation will be displayed instead of following the display unobtainable filter.
--			Changed over to using Grail's StatusCode() vice Status(), and making use of more modern bit values, thereby requiring version 20.
--			Removed a few event types that are handled because Grail now does that.  Instead switched to using Grail's new Status notification.
--			The tooltips for quests in the panel show profession and reputation requirements as appropriate.
--			Corrected a problem where the quest panel may not update properly while in combat.
--		011	Made it so the breadcrumb warning will disappear properly when the user dismisses the quest frame.
--			Made it so Grail's support for Monks does not cause headaches when Monks are not available in the game.
--			Made it so Classes that do not have any class quests will not show up in the list.
--			Put in a feature to limit quests shown to those that count towards Loremaster, thereby requiring Grail version 21.
--			When the quest details appear the quest ID is shown in the top right, and it has a tooltip with useful quest information.
--			Changed the behavior of right-clicking a quest in the quest panel to put arrows to the turn in locations for all but prerequisite quests.
--			The tooltip information describing the quest shows failure reasons by changing to red categories that fail, and to orange categories that fail in prerequisite quests.
--			The quest tooltip information now indicates the first quest(s) in the prerequisite quest list as appropriate.
--			The preference to control displaying prerequisite quests in the tooltip has been removed.
--		012	Added the ability for the tooltip to display faction reputation changes that happen when a quest is turned in.
--			Grouped the Dungeons menu items under a single Dungeons menu item.
--			Added menu items for Reputation Changes quests, grouped by reputation.
--			Added menu items for Achievement quests, grouped by continent, requiring Grail 22.
--			Updated the TOC to support Interface 40300.
--			Fixed the map pin icons whose height Blizzard changed with the 4.3 release.
--		013	Fixes a problem where map pins would not appear and the quest ID would not appear in the Blizzard Quest Frame because the events were not set up properly because sometimes Blizzard sends events in a different order than expected.
--			Makes all the Wholly quest panel update calls ensure they are performed out of combat.
--			Updates Portuguese translation thanks to weslleih and htgome.
--			Fixes a problem where quests in the Blizzard log sometimes would not appear purple in the Wholly Quest Log.
--			Fixes a problem where holidays are not detected properly because of timing issues.
--		014	Fixes the problem where the NPC tooltip did not show the exclamation point properly (instead showing a different icon) when the NPC can start a quest.
--			Adds a search ability that allows searching for quests based on their titles.
--			Adds the ability to display player coordinates into a LibDataBroker feed.
--			Updates some localizations.
--			Fixes the problem where the panel would no longer update after a UI reload, requiring Grail 26.
--			Adds some more achievements to the menu that are world event related.
--			Makes it so quests in the Blizzard quest log will be colored purple in preference to other colors (like brown in case the player would no longer qualify getting the quest).
--			Makes it so the indicator for a completed repeatable quest will appear even if the quest is not colored blue.
--		015	Adds the filtered and total quest counts to the tooltip that tells the counts of the types of quests.  For the world map button tooltip the filtered quest count displays in red if the quest markers on the map are hidden.
--			Corrects a problem where lacking LibDataBroker would cause a LUA error associated with the player coordinates.
--			Fixes a cosmetic issue with the icon in the top left of the Wholly quest log panel to show the surrounding edge properly.
--			Changes the world map check box into a button that performs the same function.
--			Changes the classification of "weekly", "monthly" and "yearly" quests so they no longer appear as resettable quests, but as normal ones.
--			Adds a tooltip for the coordinates that shows the map area ID and name.
--		016	*** Requires Grail 28 or later ***
--			Adds the ability to color the achievement menu items based on whether they are complete.
--			Corrects the problem where the tooltip does not show the different names of the NPCs that can drop an item to start a quest.
--			Corrects the problem where alt-clicking a quest would not select the NPC properly if the NPC drops an item to start a quest.
--			Tracks multiple waypoints that are logically added as a group so when one is removed all of them are removed.
--			Updates some Portuguese localizations.
--			Adds the ability to show bugged information about a quest.
--			Adds a preference to consider bugged quests unobtainable.
--			Makes it select the closest waypoint when more than one is added at the same time.
--		017 *** Requires Grail 29 or later ***
--			Updates the preferences to allow more precise control of displayed quest types.
--			Creates the ability to control whether achievement and reputation change information is used.
--			Adds some Russian localization by freega3 but abused by me.
--          Adds basic structural support for the Italian localization.
--			Changes the presentation of prerequisite quest information to have all types unified in one location.
--		018	Adds some missing Italian UI keys.
--			Removes UI keys no longer used.
--			Fixes the icon that appears in the tooltip when an NPC drops an item that starts a quest.
--			Adds the ability to display item quest prerequisites.
--			Changes the priority of quest classification to ensure truly repeatable quests are never green.
--			Adds support for Cooking and Fishing achievements, present in Grail 31.
--			Adds support to display LightHeaded data by shift-left-click a quest in the Wholly quest panel.
--			Adds the ability to display abandoned and accepted quest prerequisites.
--		019	Adds German localization provided by polzi and aryl2mithril.
--			Adds French localization provided by deadse and Noeudtribal.
--			Corrects the problem where the preference to control holiday quests always was not working properly, requiring Grail 32.
--			Updates Russian localization provided by dartraiden.
--			Adds support for Just Another Day in Tol Barad achievements when Grail provides that data (starting in Grail 32).
--			Adds the ability to display all quests from the search menu.
--			Updates Portuguese localization provided by andrewalves.
--			Corrects a rare problem interacting with LDB.
--			Adds the ability to display quest prerequisites filtering through flag quests when Grail provides the functionality.
--		020	*** Requires Grail 33 or later ***
--			Corrects the problem where quests in the log that are no longer obtainable do not appear properly.
--			Adds the ability to show daily quests that are too high for the character as orange.
--			Adds Spanish localization provided by Trisquite.
--			Moves the Daily quests into the Other category.
--			Adds the experimental option to have a wide quest panel.
--		021 *** Requires Grail 34 or later ***
--			Makes it so Mists of Pandaria reputations can be handled.
--			Makes it so starter Pandarens no longer cause LUA errors.
--			Corrects the problem where removing all TomTom waypoints was not clearing them from Wholly's memory.
--			Corrects locations for Wholly informational frames placed on QuestFrame in MoP beta.
--			Updates the tooltip to better indicate when breadcrumb quests are problems for unobtainable quests.
--			Adds the ability to display profession prerequisites (in the prerequisites section vice its own for the few that need it).
--		022	*** Requires Grail 36 or later ***
--			Corrects the problem where NPC tooltips may not be updated until the world map is shown.
--			Changes how map pins are created so no work is done unless the WorldMapFrame is being shown.
--			Adds the ability to show that quests are Scenario or Legendary.
--			Changes the artwork on the right side of the wide panel.
--			Fixes the problem where the search panel was not attaching itself to the Wholly quest panel.
--			Updates some Korean localization provided by next96.
--			Makes it so Legendary quests appear orange while daily quests that are too high level appear dark blue.
--			Adds two new sort techniques, and also a tooltip for the sort button that describes the active sort technique.
--			Adds the ability to show an abbreviated quest count for each map area in the second scroll area of the wide quest panel, with optional live updates.
--			Fixes the problem where the Wholly world map button can appear above other frames.
--			Makes changing the selection in the first scroll view in the wide version of the Wholly quest panel, remove the selection in the second scroll view, thereby allowing the zone button to properly switch to the current zone.
--			Adds a Wholly quest tooltip for each of the quests in the Blizzard quest log.
--			Updates searching in the wide frame to select the newly sought term.
--
--	Known Issues
--
--	UTF-8 file
--

local directoryName, _ = ...
local versionFromToc = GetAddOnMetadata(directoryName, "Version")
local _, _, versionValueFromToc = strfind(versionFromToc, "(%d+)")
local Wholly_File_Version = tonumber(versionValueFromToc)
local requiredGrailVersion = 36

--	Set up the bindings to use the localized name Blizzard supplies.  Note that the Bindings.xml file cannot
--	just contain the TOGGLEQUESTLOG because then the entry for Wholly does not show up.  So, we use a version
--	named WHOLLY_TOGGLEQUESTLOG which maps to the same Global string, which works exactly as we want.
BINDING_NAME_WHOLLY_TOGGLEQUESTLOG = BINDING_NAME_TOGGLEQUESTLOG
BINDING_HEADER_WHOLLY = "Wholly"

if nil == Wholly or Wholly.versionNumber < Wholly_File_Version then

	WhollyDatabase = {}

	Wholly = {

		cachedMapCounts = {},
		cachedPanelQuests = {},		-- quests and their status for map area currentPanelMapID
		cachedPinQuests = {},		-- quests and their status for map area currentPinMapID 
		checkedGrailVersion = false,	-- used so the actual check can be simpler
		checkedNPCs = {},
		checkingNPCTechniqueNew = true,
		chooseClosestWaypoint = true,
		color = {
			['B'] = "FF996600",	-- brown	[unobtainable]
			['C'] = "FF00FF00",	-- green	[completed]
			['D'] = "FF0099CC",	-- daily	[repeatable]
			['G'] = "FFFFFF00",	-- yellow	[can accept]
			['H'] = "FF0000FF", -- blue		[daily + too high level]
			['I'] = "FFFF00FF",	-- purple	[in log]
			['L'] = "FFFFFFFF",	-- white	[too high level]
			['P'] = "FFFF0000",	-- red		[does not meet prerequisites]
			['R'] = "FF0099CC",	-- daily	[true repeatable - used for question mark in pins]
			['U'] = "FF00FFFF",	-- bogus default[unknown]
			['W'] = "FF666666",	-- grey		[low-level can accept]
			['Y'] = "FFCC6600", -- orange	[legendary]
			},
		configurationScript1 = function(self)
									Wholly:ScrollFrame_Update_WithCombatCheck()
--									Wholly:UpdatePins()
									Wholly.pinsNeedFiltering = true
									Wholly:_UpdatePins()
								end,
		configurationScript2 = function(self)
--									Wholly:UpdatePins()
									Wholly:_UpdatePins()
									if Wholly.tooltip:IsVisible() and Wholly.tooltip:GetOwner() == Wholly.mapFrame then
										Wholly.tooltip:ClearLines()
										Wholly.tooltip:AddLine(Wholly.mapCountLine)
									end
								end,
		configurationScript3 = function(self)
									Wholly:_DisplayMapFrame(self:GetChecked())
								end,
		configurationScript4 = function(self)
									Wholly:UpdateQuestCaches(true)
									Wholly:ScrollFrame_Update_WithCombatCheck()
--									Wholly:UpdatePins()
									Wholly:_UpdatePins(true)
								end,
		configurationScript5 = function(self)
									Wholly:UpdateBreadcrumb(Wholly)
								end,
		configurationScript6 = function(self)
									Wholly._ForcePanelMapArea(Wholly)
									Wholly:ScrollFrame_Update_WithCombatCheck()
								end,
		configurationScript7 = function(self)
									Wholly:ScrollFrame_Update_WithCombatCheck()
								end,
		configurationScript8 = function(self)
									Wholly:UpdateCoordinateSystem()
								end,
		configurationScript9 = function(self)
									UIParentLoadAddOn("Grail-Achievements")
									Wholly:_InitializeLevelOneData()
								end,
		configurationScript10 = function(self)
									UIParentLoadAddOn("Grail-Reputations")
									Wholly:_InitializeLevelOneData()
								end,
		configurationScript11 = function(self)
									Wholly:ToggleCurrentFrame()
								end,
		configurationScript12 = function(self)
									Wholly:ScrollFrameTwo_Update()
								end,
		configurationScript13 = function(self)
								end,
		configuration = {
			[1]={[0]='Base Quests'},
			[2]={[0]='|cFF00FF00Completed|r',[1]='showsCompletedQuests',[2]='configurationScript1'},
			[3]={[0]='|cFFFF0000Needs prerequisites|r',[1]='showsQuestsThatFailPrerequsites',[2]='configurationScript1',[3]=true},
			[4]={[0]='|cFF996600Unobtainable|r',[1]='showsUnobtainableQuests',[2]='configurationScript1',[3]=true},
			[5]={[0]=FILTERS},
			[6]={[0]='|cFF0099CCRepeatable|r',[1]='showsRepeatableQuests',[2]='configurationScript1'},
			[7]={[0]='|cFF0099CC'..DAILY..'|r',[1]='showsDailyQuests',[2]='configurationScript1',[3]=true},
			[8]={[0]='|cFFFF00FFIn Log|r',[1]='showsQuestsInLog',[2]='configurationScript1',[3]=true},
			[9]={[0]='|cFF666666Low-level|r',[1]='showsLowLevelQuests',[2]='configurationScript1'},
			[10]={[0]='High level',[1]='showsHighLevelQuests',[2]='configurationScript1',[3]=true},
			[11]={[0]='Quest Title Appearance'},
			[12]={[0]="Prepend quest level",[1]='prependsQuestLevel',[2]='configurationScript1'},
			[13]={[0]="Append required level",[1]='appendRequiredLevel',[2]='configurationScript1'},
			[14]={[0]="Show whether repeatable quests previously completed",[1]='showsAnyPreviousRepeatableCompletions',[2]='configurationScript1'},
			[15]={[0]="Show status of quests in log", [1]='showsInLogQuestStatus',[2]='configurationScript7'},
			[16]={[0]=WORLD_MAP},
			[17]= {[0]="Display map pins for quest givers", [1]='displaysMapPins',[2]='configurationScript2',[4]='pairedConfigurationButton'},
			[18]= {[0]="Display button on world map", [1]='displaysMapFrame',[2]='configurationScript3'},
			[19]= {[0]="Display dungeon quests in outer map", [1]='displaysDungeonQuests',[2]='configurationScript4'},
			[20]= {[0]="Open world map updates when zones change", [1]='updatesWorldMapOnZoneChange',[2]='configurationScript1'},
			[21]={[0]='Other'},
			[22]= {[0]="Quest log panel updates when zones change", [1]='updatesPanelWhenZoneChanges',[2]='configurationScript1'},
			[23]= {[0]="Display breadcrumb quest information on Quest Frame", [1]='displaysBreadcrumbs',[2]='configurationScript5'},
			[24]= {[0]="Display holiday quests always", [1]='displaysHolidaysAlways',[2]='configurationScript6'},
			[25]= {[0]="Show only Loremaster quests", [1]='showsLoremasterOnly',[2]='configurationScript4'},
			[26]= {[0]="Enable player coordinates", [1]='enablesPlayerCoordinates',[2]='configurationScript8',[4]='pairedCoordinatesButton'},
			[27]= {[0]="Show achievement completion colors", [1]='showsAchievementCompletionColors',[2]='configurationScript1'},
			[28]= {[0]="Bugged quests considered unobtainable", [1]='buggedQuestsConsideredUnobtainable',[2]='configurationScript4'},
			[29]={[0]='Wide Wholly Quest Panel'},
			[30]={[0]='Show', [1]='useWidePanel',[2]='configurationScript11'},
			[31]={[0]='Show quest counts', [1]='showQuestCounts', [2]='configurationScript12', [3]=true},
			[32]={[0]='Live quest count updates', [1]='liveQuestCountUpdates', [2]='configurationScript13', [3]=true},
			[33]={[0]='Load Data'},
			[34]={[0]=ACHIEVEMENTS,[1]='loadAchievementData',[2]='configurationScript9'},
			[35]={[0]=COMBAT_TEXT_SHOW_REPUTATION_TEXT,[1]='loadReputationData',[2]='configurationScript10',[3]=true},
			},
		coordinates = nil,
		currentFrame = nil,
		currentPanelMapID = nil,	-- the map area the panel is showing
		currentPinMapID = nil,		-- the map area the world map is showing
		currentPinMapLevel = 0,
		debug = true,
		displaysBlizzardQuestTooltips = true,
		dropdown = nil,
		dungeonTest = {},
		eventDispatch = {
			['PLAYER_REGEN_ENABLED'] = function(self, frame)
				self:ScrollFrame_Update()
				frame:UnregisterEvent("PLAYER_REGEN_ENABLED")
			end,
			--	So in Blizzard's infinite wisdom it turns out that normal quests that just appear with the
			--	quest giver post a QUEST_DETAIL event, unless they are quests like the Candy Bucket quests
			--	which post a QUEST_COMPLETE event (even though they really are not complete until they are
			--	accepted).  And if there are more than one quest then QUEST_GREETING is posted, which also
			--	is posted if one were to decline one of the selected ones to return to the multiple choice
			--	frame again.  Therefore, it seems three events are required to ensure the breadcrumb panel
			--	is properly removed.
			['QUEST_COMPLETE'] = function(self, frame)
				local questId = self:_BreadcrumbQuestId()
				com_mithrandir_whollyQuestInfoFrameText:SetText(questId)
				self:UpdateBuggedText(questId)
				com_mithrandir_whollyBreadcrumbFrame:Hide()
			end,
			['QUEST_DETAIL'] = function(self, frame)
				local questId = self:_BreadcrumbQuestId()
				com_mithrandir_whollyQuestInfoFrameText:SetText(questId)
				self:UpdateBuggedText(questId)
				self:ShowBreadcrumbInfo()
			end,
			['QUEST_GREETING'] = function(self, frame)
				com_mithrandir_whollyQuestInfoFrameText:SetText("")
				com_mithrandir_whollyQuestInfoBuggedFrameText:SetText("")
				com_mithrandir_whollyBreadcrumbFrame:Hide()
			end,
			['QUEST_PROGRESS'] = function(self, frame)
				local questId = self:_BreadcrumbQuestId()
				com_mithrandir_whollyQuestInfoFrameText:SetText(questId)
				self:UpdateBuggedText(questId)
				com_mithrandir_whollyBreadcrumbFrame:Hide()
			end,
			['ADDON_LOADED'] = function(self, frame, arg1)
				if "Wholly" == arg1 then
					if nil == WhollyDatabase.defaultsLoaded then
						self:_LoadDefaults()
					end
					if nil == WhollyDatabase.currentSortingMode then
						WhollyDatabase.currentSortingMode = 1
					end
					if nil == WhollyDatabase.closedHeaders then
						WhollyDatabase.closedHeaders = {}
					end

					-- Now to be nicer to those that have used the addon before the current
					-- incarnation, newly added defaults will have their normal setting set
					-- as appropriate.
					if nil == WhollyDatabase.version then		-- first loaded prior to version 006, so default options added in 006
						WhollyDatabase.displaysHolidaysAlways = true		-- version 006
						WhollyDatabase.updatesWorldMapOnZoneChange = true	-- version 006
					elseif WhollyDatabase.version < 7 then
						WhollyDatabase.showsInLogQuestStatus = true			-- version 007
					elseif WhollyDatabase.version < 16 then
						WhollyDatabase.showsAchievementCompletionColors = true	-- version 016
					elseif WhollyDatabase.version < 17 then
						-- transform old values into new ones as appropriate
						if WhollyDatabase.showsDailyQuests then
							WhollyDatabase.showsRepeatableQuests = true
						end
						WhollyDatabase.loadAchievementData = true
						WhollyDatabase.loadReputationData = true
					end
					WhollyDatabase.version = Wholly.versionNumber

					self:_DisplayMapFrame(WhollyDatabase.displaysMapFrame)
					Grail:RegisterObserver("Status", self._CallbackHandler)
					Grail:RegisterObserverQuestAbandon(self._CallbackHandler)

					-- Find out which "map area" is for the player's class
					for key, value in pairs(Grail.classMapping) do
						if Grail.playerClass == value then
							self.playerClassMap = Grail.classToMapAreaMapping['C'..key]
						end
					end

					self:UpdateCoordinateSystem()	-- installs OnUpdate script appropriately

					frame:RegisterEvent("QUEST_COMPLETE")			-- to clear the breadcrumb frame
					frame:RegisterEvent("QUEST_GREETING")			-- to clear the breadcrumb frame
					frame:RegisterEvent("QUEST_PROGRESS")
					frame:RegisterEvent("WORLD_MAP_UPDATE")			-- this is for pins
					frame:RegisterEvent("ZONE_CHANGED_NEW_AREA")	-- this is for the panel
					self:UpdateBreadcrumb()							-- sets up registration of events for breadcrumbs based on user preferences
					self:_ReinstateDirectionalArrows()

					if WhollyDatabase.loadAchievementData then
						self.configurationScript9()
					end
					if WhollyDatabase.loadReputationData then
						self.configurationScript10()
					end

					-- We steal the TomTom:RemoveWaypoint() function because we want to override it ourselves
					if TomTom and TomTom.RemoveWaypoint then
						self.removeWaypointFunction = TomTom.RemoveWaypoint
						TomTom.RemoveWaypoint = function(self, uid)
							Wholly:_RemoveDirectionalArrows(uid)
							Wholly.removeWaypointFunction(TomTom, uid)
						end
					end
					if TomTom and TomTom.ClearAllWaypoints then
						self.clearAllWaypointsFunction = TomTom.ClearAllWaypoints
						TomTom.ClearAllWaypoints = function(self)
							Wholly:_RemoveAllDirectionalArrows()
							Wholly.clearAllWaypointsFunction(TomTom)
						end
					end

					self:_InitializeLevelOneData()
					if WhollyDatabase.useWidePanel then self:ToggleCurrentFrame() end

					local f = com_mithrandir_whollyFrame
					self.dropdown = CreateFrame("Button", f:GetName().."ZoneButton", f, "UIDropDownMenuTemplate")
					UIDropDownMenu_Initialize(self.dropdown, self.Dropdown_Initialize) -- took away "MENU" because no show with it
					self.dropdown:SetPoint("TOPLEFT", f, "TOPLEFT", 60, -40)
					UIDropDownMenu_SetWidth(self.dropdown, 240, 0)
					UIDropDownMenu_JustifyText(self.dropdown, "LEFT")
					-- By default, the dropdown has it clicking work with the little button on the right.  This makes it work for the whole button:
					self.dropdown:SetScript("OnClick", function(self) ToggleDropDownMenu(nil, nil, Wholly.dropdown) PlaySound("igMainMenuOptionCheckBoxOn") end)

				end
			end,
			['WORLD_MAP_UPDATE'] = function(self, frame)
--				self:UpdateQuestCaches(false, true)	-- note that we do not force the update since this event happens way too many times
				self:_UpdatePins()
			end,
			['ZONE_CHANGED_NEW_AREA'] = function(self, frame)
				if WhollyDatabase.updatesWorldMapOnZoneChange and WorldMapFrame:IsVisible() then
					local zoneNameToUse = GetRealZoneText()
					if nil ~= zoneNameToUse then
						local newMapId = Grail.zoneNameMapping[zoneNameToUse]
						if nil ~= newMapId then
							SetMapByID(newMapId)
						end
					end
				end
				self:UpdateQuestCaches(false, false, WhollyDatabase.updatesPanelWhenZoneChanges)

				if self.checkingNPCTechniqueNew then
					--	When first entering a zone for the first time the NPCs need to be studied to see whether their
					--	tooltips need to be modified with quest information.
					local newMapId = GetCurrentMapAreaID()
					if not self.checkedNPCs[newMapId] then
						self:_RecordTooltipNPCs(newMapId)
					end
				end

				-- Now update open tooltips showing our quest count data
				if GameTooltip:IsVisible() and GameTooltip:GetOwner() == com_mithrandir_whollyFrameSwitchZoneButton then
					GameTooltip:ClearLines()
					GameTooltip:AddLine(Wholly.panelCountLine)
				elseif self.tooltip:IsVisible() and self.tooltip:GetOwner() == self.mapFrame then
					self.tooltip:ClearLines()
					self.tooltip:AddLine(Wholly.mapCountLine)
				elseif GameTooltip:IsVisible() and GameTooltip:GetOwner() == self.ldbTooltipOwner then -- LibDataBroker tooltip
					GameTooltip:ClearLines()
					GameTooltip:AddLine("Wholly - " .. UIDropDownMenu_GetText(Wholly.dropdown) )
					GameTooltip:AddLine(Wholly.panelCountLine)
				elseif GameTooltip:IsVisible() and GameTooltip:GetOwner() == self.ldbCoordinatesTooltipOwner then -- LibDataBroker coordinates tooltip
					GameTooltip:ClearLines()
					local dungeonLevel = GetCurrentMapDungeonLevel()
					local dungeonIndicator = (dungeonLevel > 0) and "["..dungeonLevel.."]" or ""
					local mapAreaId = GetCurrentMapAreaID()
					local mapAreaName = Grail.mapAreaMapping[mapAreaId] or "UNKNOWN"
					GameTooltip:AddLine(string.format("%d%s %s", mapAreaId, dungeonIndicator, mapAreaName))
				end
			end,
			},
		filteredPanelQuests = {},	-- filtered table from cachedPanelQuests using current panel filters
		filteredPinQuests = {},		-- filtered table from cachedPinQuests using current pin filters
		initialUpdateProcessed = false,
		lastWhich = nil,
		lastPrerequisiteQuest = nil,
		lastUpdate = 0,
		ldbCoordinatesTooltipOwner = nil,
		ldbTooltipOwner = nil,
		levelOneCurrent = nil,
		levelOneData = nil,
		levelTwoCurrent = nil,
		levelTwoData = nil,
		mapFrame = nil,			-- the world map frame that contains the checkbox to toggle pins
		maximumSearchHistory = 10,
		npcs = {},
		pairedConfigurationButton = nil,-- configuration panel button that does the same thing as the world map button
		pairedCoordinatesButton = nil,	-- configuration panel button that does the same thing as the LDB coordinate button
		panelCountLine = "",
		pinsDisplayedLast = nil,
		pinsNeedFiltering = false,
		playerAliveReceived = false,
		playerClassMap = nil,
		previousX = 0,
		previousY = 0,
		receivedCalendarUpdateEventList = false,
		regFrames = {},
		regNPCPins = {},		-- in unifying the pins we keep a table of pins associated with NPC id
		regPins = {},
		removeWaypointFunction = nil,
		reputationLevelMapping = { [0] = 1, [35999] = 2, [38999] = 3, [41999] = 4, [44999] = 5, [50999] = 6, [62999] = 7, [83999] = 8, [84998] = 8, },
		s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['IS_BREADCRUMB'] = "Is breadcrumb quest for:",
			['PREREQUISITES'] = "Prerequisites:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "World Events",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Grail does not have this quest|r",
			['QUEST_ID'] = "Quest ID: ",
			['REQUIRED_LEVEL'] = "Required Level",
			['MAXIMUM_LEVEL_NONE'] = "None",
			['QUEST_TYPE_NORMAL'] = "Normal",
			['MAPAREA_NONE'] = "None",
			['LOREMASTER_AREA'] = "Loremaster Area",
			['FACTION_BOTH'] = "Both",
			['CLASS_NONE'] = "None",
			['CLASS_ANY'] = "Any",
			['GENDER_NONE'] = "None",
			['GENDER_BOTH'] = "Both",
			['GENDER'] = "Gender",
			['RACE_NONE'] = "None",
			['RACE_ANY'] = "Any",
			['HOLIDAYS_ONLY'] = "Available only during Holidays:",
			['SP_MESSAGE'] = "Special quest never enters Blizzard quest log",
			['INVALIDATE'] = "Invalidated by Quests:",
			['OAC'] = "On acceptance complete quests:",
			['OCC'] = "On completion of requirements complete quests:",
			['OTC'] = "On turn in complete quests:",
			['ENTER_ZONE'] = "Accepted when entering map area",
			['WHEN_KILL'] = "Accepted when killing:",
			['SEARCH_NEW'] = "New",
			['SEARCH_CLEAR'] = "Clear",
			['SEARCH_ALL_QUESTS'] = "All quests",
			['NEAR'] = "Near",
			['FIRST_PREREQUISITE'] = "First in Prerequisite Chain:",
			['BUGGED'] = "|cffff0000*** BUGGED ***|r",
			['IN_LOG'] = "In Log",
			['TURNED_IN'] = "Turned in",
			['EVER_COMPLETED'] = "Has ever been completed",
			['ITEM'] = "Item",
			['ITEM_LACK'] = "Item lack",
			['ABANDONED'] = "Abandoned",
			['NEVER_ABANDONED'] = "Never Abandoned",
			['ACCEPTED'] = "Accepted",
			['LEGENDARY'] = "Legendary",
			},
		tooltip = nil,
		updateDelay = 0.5,
		updateThreshold = 0.1,
		versionNumber = Wholly_File_Version,
		waypoints = {},

		_AchievementName = function(self, mapID)
			local colorStart, colorEnd = "", ""
			local baseName = Grail.mapAreaMapping[mapID]
			if WhollyDatabase.showsAchievementCompletionColors then
				local completed = Grail:AchievementComplete(mapID - Grail.mapAreaBaseAchievement)
				colorStart = completed and "|cff00ff00" or "|cffffff00"
				colorEnd = "|r"
			end
			return colorStart .. baseName .. colorEnd, baseName
		end,

		_AddDirectionalArrows = function(self, questTable, npcType, groupNumberToUse)
			if not TomTom or not TomTom.AddMFWaypoint then return end
			if nil == questTable or nil == npcType then return end
			local locations
			if not groupNumberToUse then
				if nil == WhollyDatabase.lastGrouping then WhollyDatabase.lastGrouping = 1 else WhollyDatabase.lastGrouping = WhollyDatabase.lastGrouping + 1 end
				if nil == WhollyDatabase.waypointGrouping then WhollyDatabase.waypointGrouping = {} end
				WhollyDatabase.waypointGrouping[WhollyDatabase.lastGrouping] = {}
			end
			for _, questId in pairs(questTable) do
				locations = Grail:QuestLocations(questId, npcType)
				if nil ~= locations then
					local indexValue = questId .. npcType
					local t = {}
					for _, npc in pairs(locations) do
						if nil ~= npc.x then
							local npcName = self:_PrettyNPCString(npc.name, npc.kill, npc.realArea)
							local uid = TomTom:AddMFWaypoint(npc.mapArea, npc.mapLevel, npc.x/100, npc.y/100,
									{	persistent = false,
										title = npcName .. " - " .. self:_QuestName(questId),
									})
							tinsert(t, uid)
						end
					end
					if 0 < #t then
						local actualGroup = groupNumberToUse or WhollyDatabase.lastGrouping
						self.waypoints[indexValue] = { grouping = actualGroup, uids = t }
						if not groupNumberToUse then
							tinsert(WhollyDatabase.waypointGrouping[WhollyDatabase.lastGrouping], indexValue)
						end
						if self.chooseClosestWaypoint and TomTom.SetClosestWaypoint and 1 < #t then
							TomTom:SetClosestWaypoint()
						end
					end
				end
			end
			if not groupNumberToUse and 0 == #(WhollyDatabase.waypointGrouping[WhollyDatabase.lastGrouping]) then
				WhollyDatabase.waypointGrouping[WhollyDatabase.lastGrouping] = nil
				WhollyDatabase.lastGrouping = WhollyDatabase.lastGrouping - 1
			end
		end,

		BreadcrumbClick = function(self, frame)
			local questId = self:_BreadcrumbQuestId()
			self:_AddDirectionalArrows(Grail:AvailableBreadcrumbs(questId), 'A')
		end,

		BreadcrumbEnter = function(self, frame)
			GameTooltip:SetOwner(frame, "ANCHOR_RIGHT")
			GameTooltip:ClearLines()
			local questId = self:_BreadcrumbQuestId()
			local breadcrumbs = Grail:AvailableBreadcrumbs(questId)
			if nil ~= breadcrumbs then
				GameTooltip:AddLine(self.s['BREADCRUMB'])
				for i = 1, #breadcrumbs do
					GameTooltip:AddLine(self:_PrettyQuestString({ breadcrumbs[i], self:_ClassificationOfQuest(breadcrumbs[i]) }))
				end
				GameTooltip:Show()
			end
		end,

		_BreadcrumbQuestId = function(self)
			local questId = GetQuestID()
			local questName = GetTitleText()

			-- Check the make sure the questId we are attempting to use makes sense with the title, otherwise
			-- the questId is incorrect and we need to try to get it
			if questName ~= self:_QuestName(questId) then
				questId = Grail:QuestIdFromNPCOrName(questName, nil, true)
			end
			return questId
		end,

		ButtonEnter = function(self, button, ...)
			local aliasQuestId = Grail:AliasQuestId(button.questId)
			local questIdToUse = aliasQuestId or button.questId
			self:_PopulateTooltipForQuest(button, questIdToUse, (questIdToUse ~= button.questId) and button.questId or nil)

			if not button.secureProcessed and not InCombatLockdown() then
				button:SetAttribute("type1", "click")
				button:SetAttribute("clickbutton", Wholly)
				button:SetAttribute("type2", "click")
				button:SetAttribute("shift-type1", "click")
				button:SetAttribute("alt-type1", "macro")
				button.secureProcessed = true
			else
				-- TODO: Should attempt a delayed setting of this if not button.secureProcessed and InCombatLockdown()
			end

			if 'P' == button.statusCode then
				self.lastPrerequisites = {}
				local lastIndexUsed = 0
				lastIndexUsed = self:_PreparePrerequisiteInfo(Grail:QuestPrerequisites(button.questId), lastIndexUsed, true)
				local outputString
				local started = false
				local tempTable = {}
				for questId, value in pairs(self.lastPrerequisites) do
					tinsert(tempTable, questId)
					outputString = ""
					if not started then
						self.tt:AddLine(" ")
						self.tt:AddLine(self.s['FIRST_PREREQUISITE'])
						started = true
					end
					for key, value2 in pairs(value) do
						if "" == outputString then
							outputString = "("..value2
						else
							outputString = outputString..","..value2
						end
					end
					outputString = outputString..") "
					self.tt:AddDoubleLine(outputString..self:_PrettyQuestString({ questId, self:_ClassificationOfQuest(questId) }), questId)
				end
				self.lastPrerequisites = started and tempTable or nil
			else
				self.lastPrerequisites = nil
			end

			if self.tt ~= GameTooltip then self.tt:Show() end
			GameTooltip:Show()
		end,

		ButtonPostClick = function(self, button)
			if button ~= self.clickingButton then print("Post click not from the same Pre click") end
			self.clickingButton = nil
		end,

		ButtonPreClick = function(self, button)
			self.clickingButton = button
		end,

		_CallbackHandler = function(type, questId)
			Wholly:UpdateQuestCaches(true)
			Wholly:_UpdatePins(true)
			if WhollyDatabase.showQuestCounts and WhollyDatabase.liveQuestCountUpdates then
				for mapId, ignoredCurrentString in pairs(Wholly.cachedMapCounts) do
					local questsInMap = Wholly:_ClassifyQuestsInMap(mapId) or {}
					Wholly.cachedMapCounts[mapId] = Wholly:_PrettyQuestCountString(questsInMap, nil, nil, true)
				end
				Wholly:ScrollFrameTwo_Update()
			end
		end,

		_CheckNPCTooltip = function(tooltip)
			if (not UnitIsPlayer("mouseover") or true) then
				-- check if this npc drops a quest item
				local id = Grail:GetNPCId(false, true)	-- only "mouseover" will be used
				local qs = id and Wholly.npcs[id] or nil
				if nil ~= qs then
					for _, questId in pairs(qs) do
						if Grail:CanAcceptQuest(questId) then
							local _, kindsOfNPC = Grail:IsTooltipNPC(id)
							if nil ~= kindsOfNPC then
								for i = 1, #(kindsOfNPC), 1 do
									local tooltipMessage = nil
									if kindsOfNPC[i][1] == Grail.NPC_TYPE_KILL then
										tooltipMessage = format(Wholly.s['KILL_TO_START_FORMAT'], Wholly:_QuestName(questId))
									elseif kindsOfNPC[i][1] == Grail.NPC_TYPE_DROP then
										if Wholly:_DroppedItemMatchesQuest(kindsOfNPC[i][2], questId) then
											tooltipMessage = format(Wholly.s['DROP_TO_START_FORMAT'], Grail:NPCName(kindsOfNPC[i][2]), Wholly:_QuestName(questId))
										end
									end
									if nil ~= tooltipMessage then
										local leftStr = format("|TInterface\\MINIMAP\\ObjectIcons:0:0:0:0:128:128:16:32:16:32|t %s", tooltipMessage)
										tooltip:AddLine(leftStr);
									end
								end
							end
							tooltip:Show();
						end
					end
				end
			end
		end,

		_ClassificationOfQuest = function(self, questId, shouldDisplayHolidays)
			local retval = 'U'
			local questBitMask, questTypeMask
			local code = ""
			local numeric = tonumber(questId)
			local subcode = ""
			if nil == numeric and nil ~= questId then
				code = strsub(questId, 1, 1)
				numeric = tonumber(strsub(questId, 2))
				if nil == numeric then
					subcode = strsub(questId, 2, 2)
					numeric = tonumber(strsub(questId, 3))
				end
			end
			if nil ~= numeric then
				if code == 'I' then
					if Grail:SpellPresent(numeric) then
						retval = 'C'
					else
						retval = 'G'
					end
				elseif code == 'J' then
					if Grail:AchievementComplete(numeric) then
						retval = 'C'
					else
						retval = 'G'
					end
				elseif code == 'K' then
					if Grail:ItemPresent(numeric) then
						retval = 'C'
					else
						retval = 'P'	-- to appear red
					end
				elseif code == 'L' then
					if Grail:ItemPresent(numeric) then
						retval = 'P'	-- to appear red
					else
						retval = 'C'
					end
				elseif code == 'M' then
					if Grail:HasQuestEverBeenAbandoned(numeric) then
						retval = 'C'
					else
						retval = 'P'
					end
				elseif code == 'N' then
					if Grail:HasQuestEverBeenAbandoned(numeric) then
						retval = 'P'
					else
						retval = 'C'
					end
				elseif code == 'P' then
					if Grail:ProfessionExceeds(subcode, numeric) then
						retval = 'C'
					else
						retval = 'P'
					end
				else
					questBitMask = Grail:StatusCode(numeric)
					questTypeMask = Grail:CodeType(numeric)
					if shouldDisplayHolidays then
						if bit.band(questBitMask, Grail.bitMaskHoliday) > 0 then
							questBitMask = questBitMask - Grail.bitMaskHoliday
						end
						if bit.band(questBitMask, Grail.bitMaskAncestorHoliday) > 0 then
							questBitMask = questBitMask - Grail.bitMaskAncestorHoliday
						end
					end
					if bit.band(questBitMask, Grail.bitMaskNonexistent + Grail.bitMaskError) > 0 then
						retval = 'U'
					elseif bit.band(questBitMask, Grail.bitMaskInLog) > 0 then
						retval = 'I'
					elseif bit.band(questBitMask, Grail.bitMaskQuestFailureWithAncestor
													+ (WhollyDatabase.buggedQuestsConsideredUnobtainable and Grail.bitMaskBugged or 0)
													) > 0 then
						retval = 'B'
					elseif bit.band(questBitMask, Grail.bitMaskCompleted + Grail.bitMaskRepeatable) == Grail.bitMaskCompleted then
						retval = 'C'
					elseif bit.band(questBitMask, Grail.bitMaskPrerequisites) > 0 then
						retval = 'P'
					elseif Grail:IsDaily(numeric) then	-- Grail.bitMaskResettable contains IsWeekly, IsMonthly and IsYearly, so we do not use because 	Blizzard shows yellow
						if bit.band(questBitMask, Grail.bitMaskLevelTooLow) > 0 then
							retval = 'H'
						else
							retval = 'D'
						end
					elseif bit.band(questBitMask, Grail.bitMaskRepeatable) > 0 then
						retval = 'R'
					elseif bit.band(questBitMask, Grail.bitMaskLevelTooLow) > 0 then
						retval = 'L'
					elseif bit.band(questBitMask, Grail.bitMaskLowLevel) > 0 then
						retval = 'W'
					elseif bit.band(questTypeMask, Grail.bitMaskQuestLegendary) > 0 then
						retval = 'Y'
					else
						retval = 'G'
					end
				end
			end
			return retval
		end,

		---
		--	Gets all the quests in the map area, then classifies them based on the current player.
		_ClassifyQuestsInMap = function(self, mapId)
			local retval = nil
			if nil ~= mapId and tonumber(mapId) then
				mapId = tonumber(mapId)
				local displaysHolidayQuestsAlways = false
				local showsLoremasterOnly = WhollyDatabase.showsLoremasterOnly
				if mapId >= Grail.mapAreaBaseHoliday and mapId <= Grail.mapAreaMaximumHoliday and WhollyDatabase.displaysHolidaysAlways then displaysHolidayQuestsAlways = true end
				retval = {}
				local questsInMap = Grail:QuestsInMap(mapId, WhollyDatabase.displaysDungeonQuests, showsLoremasterOnly) or {}
				for _,questId in pairs(questsInMap) do
					tinsert(retval, { questId, self:_ClassificationOfQuest(questId, displaysHolidayQuestsAlways) })
				end
			end
			return retval
		end,

		-- This is named this way with this function signature because it is called from the SecureActionButtonTemplate exactly like this.
		Click = function(self, leftOrRight)
			if IsShiftKeyDown() then
				if LightHeaded then self:ToggleLightHeaded() end
				return
			end
			if not TomTom or not TomTom.AddMFWaypoint then return end
			local button = self.clickingButton
			local questsToUse = {button.questId}
			local npcType = 'A'
			if "RightButton" == leftOrRight then
				if nil ~= self.lastPrerequisites then
					questsToUse = self.lastPrerequisites
				else
					npcType = 'T'
				end
			end
			self:_AddDirectionalArrows(questsToUse, npcType)
		end,

		ConfigFrame_OnLoad = function(self, panel)
			panel.name = "Wholly"
			panel:Hide()
			InterfaceOptions_AddCategory(panel)
			local parent = panel:GetName()
			local indentLevel
			local lineLevel = 0
			local button
			local offset
			
			if not self.checkedGrailVersion then
				local errorMessage = format(self.s['REQUIRES_FORMAT'], requiredGrailVersion)
				button = panel:CreateFontString(nil, "OVERLAY", "GameFontNormal")
				offset = -5
				indentLevel = 0
				lineLevel = lineLevel + 1
				button:SetPoint("TOPLEFT", panel, "TOPLEFT", (indentLevel * 200) + 8, (lineLevel * -20) + 10 + offset)
				button:SetText(errorMessage)
				return 
			end

			for i = 1, #self.configuration do
				if self.configuration[i][1] then
					button = CreateFrame("CheckButton", parent.."Button"..i, panel, "InterfaceOptionsCheckButtonTemplate")
					offset = 0
				else
					button = panel:CreateFontString(nil, "OVERLAY", "GameFontNormal")
					offset = -5
				end
				if self.configuration[i][3] then
					indentLevel = indentLevel + 1
				else
					indentLevel = 0
					lineLevel = lineLevel + 1
				end
				button:SetPoint("TOPLEFT", panel, "TOPLEFT", (indentLevel * 200) + 8, (lineLevel * -20) + 10 + offset)
				if self.configuration[i][1] then
					_G[button:GetName().."Text"]:SetText(self.configuration[i][0])
					button:SetScript("OnClick", function(self)
													WhollyDatabase[Wholly.configuration[i][1]] = self:GetChecked()
													Wholly[Wholly.configuration[i][2]](self)
												end)
					if nil ~= self.configuration[i][4] then
						self[self.configuration[i][4]] = button
					end
				else
					button:SetText(self.configuration[i][0])
				end
			end
		end,

		ConfigFrame_OnShow = function(self, panel)
			if not self.checkedGrailVersion then return end
			local parent = panel:GetName()
			for i = 1, #self.configuration do
				if self.configuration[i][1] then
					_G[parent.."Button"..i]:SetChecked(WhollyDatabase[self.configuration[i][1]])
				end
			end
		end,

		_DisplayMapFrame = function(self, shouldDisplay)
			if shouldDisplay then self.mapFrame:Show() else self.mapFrame:Hide() end
		end,

		_Distance = function(self, x1, y1, x2, y2)
			local distRatio = WorldMapDetailFrame:GetHeight() / WorldMapDetailFrame:GetWidth();
			return math.sqrt( (x1 - x2)^2 + ((y1 - y2)/distRatio)^2 );
		end,

		_Dropdown_AddButton = function(self, level, hasArrow, item)
			local info = UIDropDownMenu_CreateInfo()
			info.hasArrow = hasArrow
			info.notCheckable = true
			info.text = item.displayName
			info.value = item
			if not hasArrow then
				info.func = item.f		-- default to any menu provided function
				if nil == info.func then
					info.func = function()
						Wholly.currentPanelMapID = item.mapID
						Wholly._ForcePanelMapArea(Wholly)
						CloseDropDownMenus()
					end
				end
			end
			UIDropDownMenu_AddButton(info, level)
		end,

		Dropdown_Initialize = function(self, level)
			level = level or 1
			if 1 == level then
				for k, v in pairs(Wholly.levelOneData) do
					Wholly:_Dropdown_AddButton(level, true, v)
				end
			elseif 2 == level then
				local children = UIDROPDOWNMENU_MENU_VALUE["children"]
				if nil ~= children then
					for k, v in pairs(children) do
						Wholly:_Dropdown_AddButton(level, true, v)
					end
				else
					Wholly:_SetLevelOneCurrent(UIDROPDOWNMENU_MENU_VALUE)
					Wholly:_InitializeLevelTwoData()
					for k, v in pairs(Wholly.levelTwoData) do
						Wholly:_Dropdown_AddButton(level, false, v)
					end
				end
			else	-- assumption is level 3 which is the highest we have
				Wholly:_SetLevelOneCurrent(UIDROPDOWNMENU_MENU_VALUE)
				Wholly:_InitializeLevelTwoData()
				for k, v in pairs(Wholly.levelTwoData) do
					Wholly:_Dropdown_AddButton(level, false, v)
				end
			end
		end,

		_DroppedItemMatchesQuest = function(self, dropNPCCode, matchingQuestId)
			local retval = true
			local questCodes = Grail:CodesWithPrefixNPC(dropNPCCode, "Q:")
			if nil ~= questCodes then
				retval = false
				for i = 1, #questCodes do
					local quests = { strsplit(",", strsub(questCodes[i], 3)) }
					for j = 1, #quests do
						if tonumber(quests[j]) == tonumber(matchingQuestId) then retval = true end
					end
				end
			end
			return retval
		end,

		_FilterQuests = function(self, forPanel)
			local f = forPanel and self.filteredPanelQuests or self.filteredPinQuests
			f = {}
			local questsInMap = forPanel and self.cachedPanelQuests or self.cachedPinQuests
			local shouldAdd, statusCode, status

			--	We want to be able to force display of quests that are class or profession specific
			--	unless they are associated with the player.  In that case, the display of the quests
			--	obeys the same rules as the quests in a normal map area.
			local shouldForce = false
			if nil ~= self.currentPanelMapID and self.currentPanelMapID >= Grail.mapAreaBaseClass and self.currentPanelMapID <= Grail.mapAreaMaximumProfession then
				shouldForce = true
				if self.playerClassMap == self.currentPanelMapID then shouldForce = false end
				if self.currentPanelMapID >= Grail.mapAreaBaseProfession then
					for key,value in pairs(Grail.professionToMapAreaMapping) do
						if value == self.currentPanelMapID then
							local actualKey = key:sub(2, 2)
							if Grail:ProfessionExceeds(actualKey, 1) then -- indicates the profession is known
								shouldForce = false
							end
						end
					end
				end
			end
			local repuationQuest = false
			if nil ~= self.currentPanelMapID and self.currentPanelMapID > Grail.mapAreaBaseReputation and self.currentPanelMapID <= Grail.mapAreaMaximumReputation then
				reputationQuest = true
			end

			local questId
			local dealingWithHolidays = WhollyDatabase.displaysHolidaysAlways and nil ~= self.currentPanelMapID and self.currentPanelMapID >= Grail.mapAreaBaseHoliday and self.currentPanelMapID <= Grail.mapAreaMaximumHoliday and true or false
			local holidayModification = dealingWithHolidays and (Grail.bitMaskHoliday + Grail.bitMaskAncestorHoliday) or 0
			for i = 1, #questsInMap do
				statusCode = questsInMap[i][2]

				questId = questsInMap[i][1]
				status = Grail:StatusCode(questId)
				shouldAdd = false

				if Grail:CanAcceptQuest(questId, false, WhollyDatabase.showsQuestsThatFailPrerequsites, true, true, dealingWithHolidays) or
					(WhollyDatabase.showsCompletedQuests and Grail:IsQuestCompleted(questId) and 0 == bit.band(status, Grail.bitMaskQuestFailureWithAncestor - (Grail.bitMaskAncestorReputation + Grail.bitMaskReputation) - holidayModification)) or
					Grail:IsQuestInQuestLog(questId) or
					(WhollyDatabase.showsUnobtainableQuests and bit.band(status, Grail.bitMaskQuestFailureWithAncestor - holidayModification) > 0) then
					shouldAdd = true
				end

				if Grail:IsRepeatable(questId) and not WhollyDatabase.showsRepeatableQuests then shouldAdd = false end
				if Grail:IsDaily(questId) and not WhollyDatabase.showsDailyQuests then shouldAdd = false end
				if Grail:IsQuestInQuestLog(questId) and not WhollyDatabase.showsQuestsInLog then shouldAdd = false end
				if Grail:IsLowLevel(questId) and not WhollyDatabase.showsLowLevelQuests then shouldAdd = false end
				if bit.band(status, Grail.bitMaskLevelTooLow) > 0 and not WhollyDatabase.showsHighLevelQuests then shouldAdd = false end

				if not forPanel then
					if 'I' == statusCode or 'C' == statusCode then shouldAdd = false end
					if 'B' == statusCode then shouldAdd = false end
				end

				if shouldAdd then
					tinsert(f, questsInMap[i])
				end
			end
			if forPanel then self.filteredPanelQuests = f else self.filteredPinQuests = f end
			if not forPanel then
				self.mapCountLine = self:_PrettyQuestCountString(questsInMap, #(self.filteredPinQuests), true)
			else
				self.panelCountLine = self:_PrettyQuestCountString(questsInMap, #(self.filteredPanelQuests))
				if self.currentPanelMapID and 0 ~= self.currentPanelMapID then
					self.cachedMapCounts[self.currentPanelMapID] = self:_PrettyQuestCountString(questsInMap, nil, nil, true)
				end
			end
		end,

		_FilterPanelQuests = function(self)
			self:_FilterQuests(true)
		end,

		_FilterPinQuests = function(self)
			self:_FilterQuests(false)
		end,

		_ForcePanelMapArea = function(self, ignoreForcingSelection)
			local mapAreaName = Grail.mapAreaMapping[self.currentPanelMapID] or GetRealZoneText()	-- default to something if we do not support the zone
			if nil ~= mapAreaName then UIDropDownMenu_SetText(self.dropdown, mapAreaName) end
			self.cachedPanelQuests = self:_ClassifyQuestsInMap(self.currentPanelMapID) or {}
			self:ScrollFrame_Update_WithCombatCheck()

			if not ignoreForcingSelection then
				local soughtIndex = Grail.mapToContinentMapping[self.currentPanelMapID]
				if nil == soughtIndex then		-- assume it is a dungeon
					for i = 1, #(Grail.continents) do
						if tContains(Grail.continents[i].dungeons, self.currentPanelMapID) then
							soughtIndex = 10 + i
						end
					end
				end
				if nil == soughtIndex then		-- assume it is "Other"
					if tContains(Grail.otherMapping, self.currentPanelMapID) then
						soughtIndex = 71
					end
				end
				if nil ~= soughtIndex then
					for i, v in pairs(self.levelOneData) do
						if v.index == soughtIndex then
							self:_SetLevelOneCurrent(v)
						end
					end
				else
					self:_SetLevelOneCurrent(nil)
				end
				self:ScrollFrameOne_Update()
				--	Now we create a bogus entry for the level two data
				self:_SetLevelTwoCurrent({ displayName = mapAreaName, mapID = self.currentPanelMapID })
				self:ScrollFrameTwo_Update()
			end
		end,

		_GetMousePosition = function(self)
			local left, top = WorldMapDetailFrame:GetLeft(), WorldMapDetailFrame:GetTop();
			local width, height = WorldMapDetailFrame:GetWidth(), WorldMapDetailFrame:GetHeight();
			local scale = WorldMapDetailFrame:GetEffectiveScale();

			local x, y = GetCursorPosition();
			local cx = (x/scale - left) / width;
			local cy = (top - y/scale) / height;
	
			return math.min(math.max(cx, 0), 1), math.min(math.max(cy, 0), 1);
		end,

		_GetPin = function(self, npcId)
			if nil ~= self.regNPCPins[npcId] then return self.regNPCPins[npcId] end		-- already have a pin for this npcId

			if #self.regFrames > 0 then
				local pin = table.remove(self.regFrames)
				pin.npcId = npcId
			end

			local pin = CreateFrame("Frame", nil, WorldMapDetailFrame);
			pin.npcId = npcId
			pin:SetWidth(16);
			pin:SetHeight(16);
			pin:EnableMouse(true);
			pin:SetScript("OnEnter", function(pin) self:ShowTooltip(pin) end)
			pin:SetScript("OnLeave", function() self:_HideTooltip() end)
			pin.SetType = function(self, texType)
				if self.texType == texType then return end -- don't need to make changes
				local width, height = 0.125, 0.125

				-- 256x64, 8x4
				self.texture:SetDesaturated(false);
				self.texture:SetVertexColor(1, 1, 1);
				if texType == "D" then
					self.texture:SetTexCoord(3*width, 4*width, 1*height, 2*height);
				elseif texType == "R" then
					self.texture:SetTexCoord(4*width, 5*width, 1*height, 2*height);
				elseif texType == "P" then
					self.texture:SetTexCoord(1*width, 2*width, 1*height, 2*height);
					self.texture:SetVertexColor(1.0, 0.0, 0.0);
				elseif texType == "Y" then
					self.texture:SetTexCoord(1*width, 2*width, 1*height, 2*height);
					self.texture:SetVertexColor(12/15, 6/15, 0.0);
				elseif texType == "H" then
					self.texture:SetTexCoord(1*width, 2*width, 1*height, 2*height);
					self.texture:SetVertexColor(0.0, 0.0, 1.0);
				elseif texType == "W" then
					-- I don't like this one, it's too hard to see.
					-- self.texture:SetTexCoord(4*width, 5*width, 3*height, 4*height); -- this is the actual dimmed quest giver icon
					self.texture:SetTexCoord(1*width, 2*width, 1*height, 2*height);
					self.texture:SetVertexColor(0.75, 0.75, 0.75);
				elseif texType == "L" then
					-- Note: not all systems can desaturate. If they don't, I don't really care.
					self.texture:SetTexCoord(1*width, 2*width, 1*height, 2*height);
					self.texture:SetDesaturated(1);
				else
					self.texture:SetTexCoord(1*width, 2*width, 1*height, 2*height);
				end

				self.texType = texType
			end

			pin.texture = pin:CreateTexture();
			pin.texture:SetTexture("Interface\\MINIMAP\\ObjectIcons.blp");
			pin.texture:SetAllPoints();
			pin:SetType("L")
			self.regNPCPins[npcId] = pin
			return pin;
		end,

		_GetPrerequisiteInfo = function(self, questId, index)
			local numeric = tonumber(questId)
			if nil == numeric then numeric = tonumber(strsub(questId, 2)) end
			local statusLetter = self:_ClassificationOfQuest(questId)
			questId = numeric
			if 'P' == statusLetter then
				self:_PreparePrerequisiteInfo(Grail:QuestPrerequisites(questId), index, false)
			elseif 'U' == statusLetter or 'B' == statusLetter or 'C' == statusLetter then -- or 'L' == statusLetter
				-- do nothing since this is a failure
			else	-- I, D, R, G, W
				if nil == self.lastPrerequisites[questId] then self.lastPrerequisites[questId] = {} end
				if not tContains(self.lastPrerequisites[questId], index) then tinsert(self.lastPrerequisites[questId], index) end
			end
		end,

		_HideAllPins = function(self)
			for i, v in pairs(self.regPins) do
				self:_HidePin(i)
			end
		end,

		_HidePin = function(self, id)
			if self.regPins[id] then
				self.regPins[id]:Hide()
				table.insert(self.regFrames, self.regPins[id])
				self.regNPCPins[self.regPins[id].npcId] = nil
				self.regPins[id] = nil
			end
		end,

		_HideTooltip = function(self)
			self.tooltip:Hide()
		end,

		--	This routine will populate the data structure self.levelOneData with all of the items
		--	that are supposed to appear in the top-level dropdown or scroller.  Note that some of
		--	the items' appearances are controlled by preferences.
		_InitializeLevelOneData = function(self)
			--	each row will contain a displayName
			--	if the row is a header row, it will contain header (which is an integer so its status can be found in saved variables)
			--		and children which is a table of rows
			--	if the row is not a header row it will contain index (which is an integer used later to populate next level data)

			local entries = {}
			local t1
			
			--	Basic continents
			for i = 1, #(Grail.continents) do
				tinsert(entries, { displayName = Grail.continents[i].name, index = i })
			end
			
			--	Dungeons
			t1 = { displayName = BUG_CATEGORY3, header = 1, children = {} }
			for i = 1, #(Grail.continents) do
				tinsert(t1.children, { displayName = Grail.continents[i].name, index = 10 + i, continent = i })
			end
			tinsert(entries, t1)

			tinsert(entries, { displayName = Wholly.s['WORLD_EVENTS'], index = 21 })
			tinsert(entries, { displayName = CLASS, index = 22 })
			tinsert(entries, { displayName = TRADE_SKILLS, index = 23 })		-- Professions
			tinsert(entries, { displayName = REPUTATION, index = 24 })

			--	Achievements
			if WhollyDatabase.loadAchievementData then
				t1 = { displayName = ACHIEVEMENTS, header = 2, children = {} }
				for i = 1, #(Grail.continents) do
					tinsert(t1.children, { displayName = Grail.continents[i].name, index = 30 + i })
				end
				local i = 0
				if nil ~= Grail.worldEventAchievements and nil ~= Grail.worldEventAchievements[Grail.playerFaction] then
					for holidayKey, _ in pairs(Grail.worldEventAchievements[Grail.playerFaction]) do
						i = i + 1
						tinsert(t1.children, { displayName = Grail.holidayMapping[holidayKey], index = 40 + i, holidayName = Grail.holidayMapping[holidayKey]})
					end
				end
				i = 0
				if nil ~= Grail.professionAchievements and nil ~= Grail.professionAchievements[Grail.playerFaction] then
					for professionKey, _ in pairs(Grail.professionAchievements[Grail.playerFaction]) do
						i = i + 1
						tinsert(t1.children, { displayName = Grail.professionMapping[professionKey], index = 50 + i, professionName = Grail.professionMapping[professionKey] })
					end
				end
				tinsert(t1.children, { displayName = Wholly.s['OTHER'], index = 60 })
				tinsert(entries, t1)
			end

			--	Reputation Changes
			if WhollyDatabase.loadReputationData then

				-- 15640 is the 5.0.1 release
				local _, release = GetBuildInfo()
				release = tonumber(release)

				t1 = { displayName = COMBAT_TEXT_SHOW_REPUTATION_TEXT, header = 3, children = {} }
				tinsert(t1.children, { displayName = EXPANSION_NAME0, index = 61 })
				tinsert(t1.children, { displayName = EXPANSION_NAME1, index = 62 })
				tinsert(t1.children, { displayName = EXPANSION_NAME2, index = 63 })
				tinsert(t1.children, { displayName = EXPANSION_NAME3, index = 64 })
				if release >= 15640 then
					tinsert(t1.children, { displayName = EXPANSION_NAME4, index = 65 })
				end
				tinsert(entries, t1)
			end

			tinsert(entries, { displayName = Wholly.s['OTHER'], index = 71 })
			tinsert(entries, { displayName = SEARCH, index = 72 })

			self.levelOneData = entries			
		end,

		--	This routine will populate the data structure self.levelTwoData with all of the items
		--	that are supposed to appear in the next-level dropdown or scroller based on the level
		--	one selection.
		_InitializeLevelTwoData = function(self)
			local t = {}
			local which = self.levelOneCurrent and self.levelOneCurrent.index or nil
			if nil == which then self.levelTwoData = t return end
			if 10 > which then				-- Basic continent
				local Z = Grail.continents[which].zones
				for i = 1, #Z do
					local t1 = {}
					t1.sortName = Z[i].name
					t1.displayName = Z[i].name
					t1.mapID = Z[i].mapID
					tinsert(t, t1)
				end
			elseif 20 > which then			-- Dungeons
				local mapAreas = Grail.continents[self.levelOneCurrent.continent].dungeons
				for i = 1, #mapAreas do
					local t1 = {}
					t1.sortName = Grail.mapAreaMapping[mapAreas[i]]
					t1.displayName = t1.sortName
					t1.mapID = mapAreas[i]
					tinsert(t, t1)
				end
			elseif 21 == which then			-- World Events
				for code, name in pairs(Grail.holidayMapping) do
					local t1 = {}
					t1.sortName = name
					t1.displayName = name
					t1.mapID = Grail.holidayToMapAreaMapping['H'..code]
					tinsert(t, t1)
				end
			elseif 22 == which then		-- Class
				for code, englishName in pairs(Grail.classMapping) do
					local localizedGenderClassName = Grail:CreateClassNameLocalizedGenderized(englishName)
					local classColor = RAID_CLASS_COLORS[englishName]
					local mapId = Grail.classToMapAreaMapping['C'..code]
					if nil == classColor then
						classColor = { r = 0.0, g = 1.0, b = 150/255 }
						localizedGenderClassName = "Monk"
					end	-- need to do for Monk currently
					if nil ~= Grail.mapAreaMapping[mapId] then
						local t1 = {}
						t1.sortName = localizedGenderClassName
						t1.displayName = format("|cff%.2x%.2x%.2x%s|r", classColor.r*255, classColor.g*255, classColor.b*255, localizedGenderClassName)
						t1.mapID = mapId
						tinsert(t, t1)
					end
				end
			elseif 23 == which then		-- Professions
				for code, professionName in pairs(Grail.professionMapping) do
					local mapId = Grail.professionToMapAreaMapping['P'..code]
					if nil ~= Grail.mapAreaMapping[mapId] then
						local t1 = {}
						t1.sortName = professionName
						t1.displayName = professionName
						t1.mapID = mapId
						tinsert(t, t1)
					end
				end
			elseif 24 == which then		-- Reputations
				for reputationIndex, reputationName in pairs(Grail.reputationMapping) do
					local factionId = tonumber(reputationIndex, 16)
					local mapId = Grail.mapAreaBaseReputation + factionId
--					local mapId = Grail.mapAreaBaseReputation + reputationIndex
					if nil ~= Grail.mapAreaMapping[mapId] then
						local t1 = {}
						t1.sortName = reputationName
						t1.displayName = reputationName
						t1.mapID = mapId
						tinsert(t, t1)
					end
				end
			elseif 40 > which then		-- Continent Achievements
				local mapAreas = Grail.achievements[Grail.playerFaction] and Grail.achievements[Grail.playerFaction][which - 30] or {}
				for i = 1, #mapAreas do
					local t1 = {}
					t1.sortName = Grail.mapAreaMapping[mapAreas[i]]
					t1.displayName = self:_AchievementName(mapAreas[i])
					t1.mapID = mapAreas[i]
					tinsert(t, t1)
				end
			elseif 50 > which then		-- Holiday Achievements
				local mapAreas = Grail.worldEventAchievements[Grail.playerFaction] and Grail.worldEventAchievements[Grail.playerFaction][Grail.reverseHolidayMapping[self.levelOneCurrent.holidayName]] or {}
				for i = 1, #mapAreas do
					local t1 = {}
					t1.sortName = Grail.mapAreaMapping[mapAreas[i]]
					t1.displayName = self:_AchievementName(mapAreas[i])
					t1.mapID = mapAreas[i]
					tinsert(t, t1)
				end
			elseif 60 > which then		-- Profession Achievements
				local mapAreas = Grail.professionAchievements[Grail.playerFaction] and Grail.professionAchievements[Grail.playerFaction][Grail.reverseProfessionMapping[self.levelOneCurrent.professionName]] or {}
				for i = 1, #mapAreas do
					local t1 = {}
					t1.sortName = Grail.mapAreaMapping[mapAreas[i]]
					t1.displayName = self:_AchievementName(mapAreas[i])
					t1.mapID = mapAreas[i]
					tinsert(t, t1)
				end
			elseif 60 == which then		-- Other Achievements
				-- 5 Dungeon Achievement
				local t1 = {}
				local mapID = Grail.mapAreaBaseAchievement + 4956
				t1.displayName, t1.sortName = self:_AchievementName(mapID)
				t1.mapID = mapID
				tinsert(t, t1)

				-- Just Another Day in Tol Barad Achievement
				t1 = {}
				mapID = Grail.mapAreaBaseAchievement + ("Alliance" == Grail.playerFaction and 5718 or 5719)
				t1.displayName, t1.sortName = self:_AchievementName(mapID)
				t1.mapID = mapID
				tinsert(t, t1)
			elseif 70 > which then		-- Reputation Changes
				local mapAreas = Grail.reputationExpansionMapping[which - 60]
				for i = 1, #mapAreas do
					local t1 = {}
					local mapID = Grail.mapAreaBaseReputationChange + mapAreas[i]
					local factionId = Grail:_HexValue(mapAreas[i], 3)
					t1.sortName = Grail.reputationMapping[factionId]
					t1.displayName = t1.sortName
					t1.mapID = mapID
					if nil ~= Grail.indexedQuests[mapID] and 0 ~= #(Grail.indexedQuests[mapID]) then
						tinsert(t, t1)
					end
				end
			elseif 71 == which then		-- Other
				for i = 1, #(Grail.otherMapping) do
					local t1 = {}
					local mapID = Grail.otherMapping[i]
					t1.sortName = Grail.mapAreaMapping[mapID]
					t1.displayName = t1.sortName
					t1.mapID = mapID
					tinsert(t, t1)
				end
				local mapAreaID = Grail.mapAreaBaseDaily
				local mapName = Grail.mapAreaMapping[mapAreaID]
				tinsert(t, { sortName = mapName, displayName = "|c" .. self.color['D'] .. mapName .. "|r", mapID = mapAreaID })
			elseif 72 == which then		-- Search
				-- We use sortName in a special way because we do not want these items sorted alphabetically
				local lastUsed = 1
				tinsert(t, { sortName = 1, displayName = Wholly.s['SEARCH_NEW'], f = function() com_mithrandir_whollySearchFrame:Show() Wholly.currentPanelMapID = nil Wholly._SetLevelTwoCurrent(Wholly, nil) Wholly._ForcePanelMapArea(Wholly,true) CloseDropDownMenus() end })
				if WhollyDatabase.searches and 0 < #(WhollyDatabase.searches) then
					for i = 1, #(WhollyDatabase.searches) do
						local shouldSelect = (i == #(WhollyDatabase.searches)) and self.justAddedSearch
						tinsert(t, { sortName = i + 1, displayName = SEARCH .. ': ' .. WhollyDatabase.searches[i], mapID = 0, selected = shouldSelect, f = function() Wholly.SearchForQuestNamesMatching(Wholly, WhollyDatabase.searches[i]) Wholly.currentPanelMapID = 0 Wholly._ForcePanelMapArea(Wholly, true) CloseDropDownMenus() end })
					end
					lastUsed = #(WhollyDatabase.searches) + 2
					tinsert(t, { sortName = lastUsed, displayName = Wholly.s['SEARCH_CLEAR'], f = function() WhollyDatabase.searches = nil CloseDropDownMenus() Wholly.currentPanelMapID = nil Wholly._SetLevelTwoCurrent(Wholly, nil) Wholly._ForcePanelMapArea(Wholly,true) Wholly.ScrollFrameTwo_Update(Wholly) end })
					self.justAddedSearch = nil
				end
				tinsert(t, { sortName = lastUsed + 1, displayName = Wholly.s['SEARCH_ALL_QUESTS'], f = function() Wholly.SearchForAllQuests(Wholly) Wholly.currentPanelMapID = 0 Wholly._ForcePanelMapArea(Wholly, true) CloseDropDownMenus() end })
			end
			table.sort(t, function(a, b) return a.sortName < b.sortName end)

			-- We want to make sure we retain the proper selection
			if nil ~= self.levelTwoCurrent then
				for i, v in pairs(t) do
					if v.displayName == self.levelTwoCurrent.displayName and v.mapID == self.levelTwoCurrent.mapID then
						v.selected = true
					end
				end
			end
			self.levelTwoData = t
			self.lastWhich = which
		end,

		_LoadDefaults = function(self)
			local db = {}
			db.defaultsLoaded = true
			db.prependsQuestLevel = true
			db.appendRequiredLevel = true
			db.showsLowLevelQuests = true
			db.showsAnyPreviousRepeatableCompletions = true
			db.updatesPanelWhenZoneChanges = true
			db.displaysMapPins = true
			db.displaysMapFrame = true
			db.displaysDungeonQuests = true
			db.displaysBreadcrumbs = true
			db.displaysHolidaysAlways = true
			db.updatesWorldMapOnZoneChange = true
			db.showsInLogQuestStatus = true
			db.showsAchievementCompletionColors = true
			db.loadAchievementData = true
			db.loadReputationData = true
			db.version = Wholly.versionNumber
			WhollyDatabase = db
		end,

		_NPCInfoSection = function(self, heading, table, button, meetsCriteria)
			if nil == table then return end
			self.tt:AddLine(" ")
			self.tt:AddLine(heading)
			for _, npcId in pairs(table) do
				local locations = Grail:NPCLocations(npcId)
				if nil ~= locations then
					for _, npc in pairs(locations) do
						local locationString = npc.mapArea and Grail.mapAreaMapping[npc.mapArea] or ""
						if npc.near then
							locationString = locationString .. ' ' .. self.s['NEAR']
						elseif nil ~= npc.x then
							locationString = locationString .. string.format(' %.2f, %.2f', npc.x, npc.y)
						end
						local nameToUse = npc.name
						if npc.dropName then
							nameToUse = nameToUse .. " (" .. npc.dropName .. ')'
						end
						self.tt:AddDoubleLine(self:_PrettyNPCString(nameToUse, npc.kill, npc.realArea), locationString)
						if meetsCriteria then
							local desiredMacroValue = SLASH_TARGET1 .. ' ' .. npc.name
							if button:GetAttribute("macrotext") ~= desiredMacroValue and not InCombatLockdown() then
								button:SetAttribute("macrotext", desiredMacroValue)
							end
						end
					end
				end
			end
		end,

		_OnEvent = function(self, frame, event, ...)
			if self.eventDispatch[event] then
				self.eventDispatch[event](self, frame, ...)
			end
		end,

		OnHide = function(self, frame)
		end,

		OnLoad = function(self, frame)
			if not Grail or Grail.versionNumber < requiredGrailVersion then
				local errorMessage = format(self.s['REQUIRES_FORMAT'], requiredGrailVersion)
				print(errorMessage)
				UIErrorsFrame:AddMessage(errorMessage)
				return 
			end
			self.checkedGrailVersion = true
			SlashCmdList["WHOLLY"] = function(msg)
				self:SlashCommand(frame, msg)
			end
			SLASH_WHOLLY1 = "/wholly"
			com_mithrandir_whollyFrameTitleText:SetText("Wholly ".. com_mithrandir_whollyFrameTitleText:GetText())
			com_mithrandir_whollyFrameWideTitleText:SetText("Wholly ".. com_mithrandir_whollyFrameWideTitleText:GetText())

			self.currentFrame = com_mithrandir_whollyFrame

			-- The frame is not allowing button presses to things just on the outside of its bounds so we move the hit rect
			frame:SetHitRectInsets(0, 32, 0, 84)

			local LibStub = _G["LibStub"]
			if LibStub then
				local LDB = LibStub("LibDataBroker-1.1", true)
				if LDB then
					local launcher = LDB:NewDataObject("Wholly", { type="launcher", icon="Interface\\Icons\\INV_Misc_Book_07",
							OnClick = function(theFrame, button) if button == "RightButton" then InterfaceOptionsFrame_OpenToCategory("Wholly") else frame:Show() end end,
							OnTooltipShow = function(tooltip)
								Wholly:_ProcessInitialUpdate()
								Wholly.ldbTooltipOwner = tooltip:GetOwner()
								local dropdownValue = UIDropDownMenu_GetText(Wholly.dropdown)
								local printValue = dropDownValue or ""
								tooltip:AddLine("Wholly - " .. printValue )
								tooltip:AddLine(Wholly.panelCountLine)
								end, 
							})
					self.coordinates = LDB:NewDataObject("Wholly Coordinates", { type="data source", icon="Interface\\Icons\\INV_Misc_Map02", text="",
							OnClick = function(theFrame, button) Wholly.pairedCoordinatesButton:Click() end,
							OnTooltipShow = function(tooltip)
								Wholly.ldbCoordinatesTooltipOwner = tooltip:GetOwner()
								local dungeonLevel = GetCurrentMapDungeonLevel()
								local dungeonIndicator = (dungeonLevel > 0) and "["..dungeonLevel.."]" or ""
								local mapAreaId = GetCurrentMapAreaID()
								local mapAreaName = Grail.mapAreaMapping[mapAreaId] or "UNKNOWN"
								tooltip:AddLine(string.format("%d%s %s", mapAreaId, dungeonIndicator, mapAreaName)) end,
							})
				end
			end

			self.tooltip = CreateFrame("GameTooltip", "com_mithrandir_WhollyTooltip", UIParent, "GameTooltipTemplate");
			self.tooltip:SetFrameStrata("TOOLTIP");
			self.tooltip.large = com_mithrandir_WhollyTooltipTextLeft1:GetFontObject();
			self.tooltip.small = com_mithrandir_WhollyTooltipTextLeft2:GetFontObject();
			self.tooltip.SetLastFont = function(self, fontObj, rightText)
				local txt = rightText and "Right" or "Left"
				_G[format("com_mithrandir_WhollyTooltipText%s%d", txt, self:NumLines())]:SetFont(fontObj:GetFont())
			end

			self.tt2 = CreateFrame("GameTooltip", "com_mithrandir_WhollyOtherTooltip", GameTooltip, "GameTooltipTemplate")

			local f = CreateFrame("Button", nil, WorldMapFrame, "UIPanelButtonTemplate")
			f:SetSize(100, 25)
			if nil == Gatherer_WorldMapDisplay then
				f:SetPoint("TOPLEFT", WorldMapPositioningGuide, "TOPLEFT", 4, -4)
			else
				f:SetPoint("TOPLEFT", Gatherer_WorldMapDisplay, "TOPRIGHT", 4, 0)
			end
			f:SetToplevel(true)
			f:SetScale(0.7)
			f:SetText("Wholly")
			f:SetScript("OnShow", function(self)
									if nil == Gatherer_WorldMapDisplay then
										if not(GetCVarBool("miniWorldMap")) then
											self:SetPoint("TOPLEFT", WorldMapPositioningGuide, "TOPLEFT", 4, -4)
										else
											self:SetPoint("TOPLEFT", WorldMapTitleButton, "TOPLEFT", 8, -3)
										end
									else
										self:SetPoint("TOPLEFT", Gatherer_WorldMapDisplay, "TOPRIGHT", 4, 0)
									end
								end)
			f:SetScript("OnEnter", function(self) local t = Wholly.tooltip t:ClearLines() t:SetOwner(self) t:AddLine(Wholly.mapCountLine) t:Show() t:ClearAllPoints() t:SetPoint("TOPLEFT", self, "BOTTOMRIGHT") end)
			f:SetScript("OnLeave", function(self) Wholly.tooltip:Hide() end)
			f:SetScript("OnClick", function(self) Wholly.pairedConfigurationButton:Click() end)
			f:Hide()
			self.mapFrame = f

			-- if the UI panel disappears (maximized WorldMapFrame) we need to change parents
			UIParent:HookScript("OnHide", function()
				self.tooltip:SetParent(WorldMapFrame);
				self.tooltip:SetFrameStrata("TOOLTIP");
			end)
			UIParent:HookScript("OnShow", function()
				self.tooltip:SetParent(UIParent);
				self.tooltip:SetFrameStrata("TOOLTIP");
			end)

			WorldMapFrame:HookScript("OnShow", function()
				Wholly:_UpdatePins()
			end)

			GameTooltip:HookScript("OnTooltipSetUnit", Wholly._CheckNPCTooltip)

			if self.displaysBlizzardQuestTooltips then
				local buttons = QuestLogScrollFrame.buttons
				local buttonCount = #buttons
				for i = 1, buttonCount do
					buttons[i]:HookScript("OnEnter", function(button)
						self:_PresentTooltipForBlizzardQuest(button)
					end)
					buttons[i]:HookScript("OnLeave", function(button)
						GameTooltip:Hide()
					end)
				end
			end

			-- Our frame positions are wrong for MoP, so we change them here.
			if Grail.blizzardRelease >= 15640 then
				com_mithrandir_whollyQuestInfoFrame:SetPoint("TOPRIGHT", QuestFrame, "TOPRIGHT", -15, -35)
				com_mithrandir_whollyQuestInfoBuggedFrame:SetPoint("TOPLEFT", QuestFrame, "TOPLEFT", 100, -35)
				com_mithrandir_whollyBreadcrumbFrame:SetPoint("TOPLEFT", QuestFrame, "BOTTOMLEFT", 16, -10)
			end

			local nf = CreateFrame("Frame")
			nf:SetScript("OnEvent", function(frame, event, ...) self:_OnEvent(frame, event, ...) end)
			nf:RegisterEvent("ADDON_LOADED")
			self.notificationFrame = nf

			if "deDE" == GetLocale() then
				com_mithrandir_whollyFramePreferencesButton:SetText("Einstellungen")
			end
			if "ruRU" == GetLocale() then
				com_mithrandir_whollyFrameSortButton:SetText("Сортировать")
			end

		end,

		---
		--	The first time the panel is shown it is populated with the information from the current map area.
		OnShow = function(self, frame)
			self:_ProcessInitialUpdate()
			if WhollyDatabase.showsInLogQuestStatus then
				self:ScrollFrame_Update_WithCombatCheck()
			end
			self:ScrollFrameOne_Update()
			self:ScrollFrameTwo_Update()
		end,

		_OnUpdate = function(self, frame, elapsed)
			self.lastUpdate = self.lastUpdate + elapsed
			if self.lastUpdate < self.updateThreshold then return end
			local x, y = GetPlayerMapPosition('player')
			if self.previousX ~= x or self.previousY ~= y then
				if nil ~= self.coordinates then
					self.coordinates.text = string.format("%.2f, %.2f", x * 100, y * 100)
				end
				self.previousX = x
				self.previousY = y
			end
			self.lastUpdate = 0
		end,

		_PresentTooltipForBlizzardQuest = function(self, blizzardQuestButton)
			local questIndex = blizzardQuestButton:GetID()
			local questTitle, level, questTag, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily, questId, startEvent = GetQuestLogTitle(questIndex)
			if not isHeader then
				self:_PopulateTooltipForQuest(blizzardQuestButton, questId)
				GameTooltip:Show()
			end
		end,

		_PrettyNPCString = function(self, npcName, mustKill, realAreaId)
			if mustKill then npcName = format(self.s['MUST_KILL_PIN_FORMAT'], npcName) end
			if realAreaId then npcName = format("%s => %s", npcName, Grail.mapAreaMapping[realAreaId]) end
			return npcName
		end,

		_PrettyQuestCountString = function(self, questTable, displayedCount, forMap, abbreviated)
			local retval = ""
			local codesToUse = abbreviated and { 'G', 'W', 'L', 'Y', } or { 'G', 'W', 'L', 'Y', 'P', 'D', 'R', 'H', 'I', 'C', 'B', }
			local lastCode = abbreviated and 'P' or 'U'
			displayedCount = displayedCount or 0
			if nil ~= questTable then
				local totals = { ['B'] = 0, ['C'] = 0, ['D'] = 0, ['G'] = 0, ['H'] = 0, ['I'] = 0, ['L'] = 0, ['P'] = 0, ['R'] = 0, ['U'] = 0, ['W'] = 0, ['Y'] = 0, }
				local code
				for i = 1, #questTable do
					code = questTable[i][2]
					totals[code] = totals[code] + 1
				end
				local colorCode
				for _,code in pairs(codesToUse) do
					colorCode = self.color[code]
					retval = retval .. "|c" .. colorCode .. totals[code] .. "|r/"
				end
				local displayStart, displayEnd = "", ""
				if forMap and not WhollyDatabase.displaysMapPins then
					displayStart, displayEnd = "|cffff0000", "|r"
				end
				retval = retval .. "|c" .. self.color[lastCode] .. totals[lastCode] .. "|r"
				if not abbreviated then
					retval = retval .. "  [" .. displayStart .. displayedCount .. displayEnd .. "/" .. #questTable .."]"
				end
			end
			return retval
		end,

		_PrettyQuestString = function(self, questTable)
			local questId = questTable[1]
			local questCode = ""
			local numeric = tonumber(questId)
			local subcode = ""
			if nil == numeric then
				questCode = strsub(questId, 1, 1)
				numeric = tonumber(strsub(questId, 2))
				if nil == numeric then
					subcode = strsub(questId, 2, 2)
					numeric = tonumber(strsub(questId, 3))
				end
			end
			local filterCode = questTable[2]
			local colorCode = self.color[filterCode]
			if questCode == 'I' then
				local name = GetSpellInfo(numeric)
				return format("|c%s%s|r [%s]", colorCode, name, SPELLS)
			elseif questCode == 'J' then
				local id, name = GetAchievementInfo(numeric)
				return format("|c%s%s|r [%s]", colorCode, name, ACHIEVEMENTS)
			elseif questCode == 'K' then
				local name = Grail:NPCName(numeric)
				return format("|c%s%s|r [%s]", colorCode, name, self.s['ITEM'])
			elseif questCode == 'L' then
				local name = Grail:NPCName(numeric)
				return format("|c%s%s|r [%s]", colorCode, name, self.s['ITEM_LACK'])
			elseif questCode == 'P' then
				local meetsRequirement, actualSkillLevel = Grail:ProfessionExceeds(subcode, numeric)
				local levelCode
				if meetsRequirement then
					colorCode = self.color['C']
					levelCode = self.color['C']
				elseif actualSkillLevel ~= Grail.NO_SKILL then
					colorCode = self.color['C']
					levelCode = self.color['P']
				else
					colorCode = self.color['P']
					levelCode = self.color['P']
				end
				return format("|c%s%s|r |c%s%d|r [%s]", colorCode, Grail.professionMapping[subcode], levelCode, numeric, TRADE_SKILLS)
			else
				questId = numeric
				local typeString = ""
				if questCode == 'B' then
					typeString = format(" [%s]", self.s['IN_LOG'])
				elseif questCode == 'C' then
					typeString = format(" [%s, %s]", self.s['IN_LOG'], self.s['TURNED_IN'])
				elseif questCode == 'D' then
					typeString = format(" [%s]", COMPLETE)
				elseif questCode == 'E' then
					typeString = format(" [%s, %s]", COMPLETE, self.s['TURNED_IN'])
				elseif questCode == 'H' then
					typeString = format(" [%s]", self.s['EVER_COMPLETED'])
				elseif questCode == 'M' then
					typeString = format(" [%s]", self.s['ABANDONED'])
				elseif questCode == 'N' then
					typeString = format(" [%s]", self.s['NEVER_ABANDONED'])
				elseif questCode == 'O' then
					typeString = format(" [%s]", self.s['ACCEPTED'])
				end
				local statusCode = Grail:StatusCode(questId)
				local questLevel = Grail:QuestLevel(questId)
				local questLevelString = WhollyDatabase.prependsQuestLevel and format("[%s] ", questLevel or "??") or ""
				local requiredLevelString = ""
				if WhollyDatabase.appendRequiredLevel then
					local success, _, questLevelNeeded, _ = Grail:MeetsRequirementLevel(questId)
					if bit.band(statusCode, Grail.bitMaskLevelTooLow) > 0 then requiredLevelString = format(" [%s]", questLevelNeeded) end
				end
				local repeatableCompletedString = WhollyDatabase.showsAnyPreviousRepeatableCompletions and bit.band(statusCode, Grail.bitMaskResettableRepeatableCompleted) > 0 and "*" or ""
				return format("|c%s%s%s%s%s|r%s", colorCode, questLevelString, self:_QuestName(questId), repeatableCompletedString, requiredLevelString, typeString)
			end
		end,

		_ProcessInitialUpdate = function(self)
			if not self.initialUpdateProcessed then
				self.currentPanelMapID = GetCurrentMapAreaID()
				self:_ForcePanelMapArea()
				self.initialUpdateProcessed = true
			end
		end,
			
		_PopulateTooltipForQuest = function(self, frame, questId, aliasQuestId)
			self.tt = GameTooltip
			questId = tonumber(questId)
			self.tt:SetOwner(frame, "ANCHOR_RIGHT")
			self.tt:ClearLines()
			if nil == questId then return end
			self.tt:SetHyperlink(format("quest:%d", questId))
			if not Grail:DoesQuestExist(questId) then self.tt:AddLine(" ") self.tt:AddLine(self.s['GRAIL_NOT_HAVE']) return end

			local bugged = Grail:IsBugged(questId)
			if bugged then
				self.tt:AddLine(" ")
				self.tt:AddLine(self.s['BUGGED'])
				self.tt:AddLine(bugged)
			end

			if self.debug then
				self.tt:AddLine(" ")
				local aliasQuestString = aliasQuestId and " ("..aliasQuestId..")" or ""
				self.tt:AddLine(self.s['QUEST_ID']..questId..aliasQuestString)
			end

			questId = aliasQuestId or questId	-- remap to the alias now that the Blizzard interaction is done
			local obtainersCode = Grail:CodeObtainers(questId)
			local holidayCode = Grail:CodeHoliday(questId)
			local questLevel = Grail:QuestLevel(questId)
			local _, _, requiredLevel, notToExceedLevel = Grail:MeetsRequirementLevel(questId)
			local questType = self:_QuestTypeString(questId):gsub("^%s*(.-)%s*$", "%1") -- remove the leading and trailing whitespace
			local statusCode = Grail:StatusCode(questId)
			local normalColor, redColor, orangeColor, greenColor = "ffffd200", "ffff0000", "ffff9900", "ff00ff00"
			local colorCode

			self.tt:AddLine(" ")
			self.tt:AddDoubleLine(LEVEL, questLevel)
			self.tt:AddDoubleLine(self.s['REQUIRED_LEVEL'], requiredLevel)
			if bit.band(statusCode, Grail.bitMaskLevelTooHigh) > 0 then colorCode = redColor elseif bit.band(statusCode, Grail.bitMaskAncestorLevelTooHigh) > 0 then colorCode = orangeColor else colorCode = normalColor end
			self.tt:AddDoubleLine("|c"..colorCode..GUILD_RECRUITMENT_MAXLEVEL.."|r", (notToExceedLevel * Grail.bitMaskQuestMaxLevelOffset == Grail.bitMaskQuestMaxLevel) and self.s['MAXIMUM_LEVEL_NONE'] or notToExceedLevel)

			if "" == questType then questType = self.s['QUEST_TYPE_NORMAL'] end
			local len = strlen(questType)
			if len > 0 and ' ' == strsub(questType, len, len) then questType = strsub(questType, 1, len - 1) end
			self.tt:AddDoubleLine(TYPE, questType)

			local loremasterString = self.s['MAPAREA_NONE']
			local loremasterMapArea = Grail:LoremasterMapArea(questId)
			if nil ~= loremasterMapArea then loremasterString = Grail.mapAreaMapping[loremasterMapArea] end
			self.tt:AddDoubleLine(self.s['LOREMASTER_AREA'], loremasterString)

			self.tt:AddLine(" ")
			local factionString = FACTION_OTHER
			if Grail.bitMaskFactionAll == bit.band(obtainersCode, Grail.bitMaskFactionAll) then
				factionString = self.s['FACTION_BOTH']
			elseif 0 < bit.band(obtainersCode, Grail.bitMaskFactionAlliance) then
				factionString = FACTION_ALLIANCE
			elseif 0 < bit.band(obtainersCode, Grail.bitMaskFactionHorde) then
				factionString = FACTION_HORDE
			end
			if bit.band(statusCode, Grail.bitMaskFaction) > 0 then colorCode = redColor elseif bit.band(statusCode, Grail.bitMaskAncestorFaction) > 0 then colorCode = orangeColor else colorCode = normalColor end
			self.tt:AddDoubleLine("|c"..colorCode..FACTION.."|r", factionString)

			local classString
			if 0 == bit.band(obtainersCode, Grail.bitMaskClassAll) then
				classString = self.s['CLASS_NONE']
			elseif Grail.bitMaskClassAll == bit.band(obtainersCode, Grail.bitMaskClassAll) then
				classString = self.s['CLASS_ANY']
			else
				classString = ""
				for letterCode, bitValue in pairs(Grail.classToBitMapping) do
					if 0 < bit.band(obtainersCode, bitValue) then
						local englishName = Grail.classMapping[letterCode]
						local localizedGenderClassName = Grail:CreateClassNameLocalizedGenderized(englishName)
						local classColor = RAID_CLASS_COLORS[englishName]
						-- TODO: Remove the following line when we actually have Monks available
						if nil == classColor then classColor = { r = 1.0, g = 1.0, b = 1.0 } localizedGenderClassName = "Monk" end
						classString = classString .. format("|cff%.2x%.2x%.2x%s|r ", classColor.r*255, classColor.g*255, classColor.b*255, localizedGenderClassName)
					end
				end
				len = strlen(classString)
				if 0 > len and ' ' == strsub(classString, len, len) then classString = strsub(classString, 1, len - 1) end
			end
			if bit.band(statusCode, Grail.bitMaskClass) > 0 then colorCode = redColor elseif bit.band(statusCode, Grail.bitMaskAncestorClass) > 0 then colorCode = orangeColor else colorCode = normalColor end
			self.tt:AddDoubleLine("|c"..colorCode..CLASS.."|r", classString)

			local genderString = self.s['GENDER_NONE']
			if Grail.bitMaskGenderAll == bit.band(obtainersCode, Grail.bitMaskGenderAll) then
				genderString = self.s['GENDER_BOTH']
			elseif 0 < bit.band(obtainersCode, Grail.bitMaskGenderMale) then
				genderString = MALE
			elseif 0 < bit.band(obtainersCode, Grail.bitMaskGenderFemale) then
				genderString = FEMALE
			end
			if bit.band(statusCode, Grail.bitMaskGender) > 0 then colorCode = redColor elseif bit.band(statusCode, Grail.bitMaskAncestorGender) > 0 then colorCode = orangeColor else colorCode = normalColor end
			self.tt:AddDoubleLine("|c"..colorCode..self.s['GENDER'].."|r", genderString)

			-- Note that race can show races of any faction, especially if the quest is marked just to exclude a specific race
			local raceString
			if 0 == bit.band(obtainersCode, Grail.bitMaskRaceAll) then
				raceString = self.s['RACE_NONE']
			elseif Grail.bitMaskRaceAll == bit.band(obtainersCode, Grail.bitMaskRaceAll) then
				raceString = self.s['RACE_ANY']
			else
				raceString = ""
				for letterCode, bitValue in pairs(Grail.raceToBitMapping) do
					if 0 < bit.band(obtainersCode, bitValue) then
						local englishName = Grail.raceMapping[letterCode]
						local localizedGenderRaceName = Grail:CreateRaceNameLocalizedGenderized(englishName)
						-- TODO: Remove the following line when we actually have Pandarens available
						if nil == localizedGenderRaceName then localizedGenderRaceName = "Pandaren" end
						raceString = raceString .. localizedGenderRaceName .. " "
					end
				end
				len = strlen(raceString)
				if 0 > len and ' ' == strsub(raceString, len, len) then raceString = strsub(raceString, 1, len - 1) end
			end
			if bit.band(statusCode, Grail.bitMaskRace) > 0 then colorCode = redColor elseif bit.band(statusCode, Grail.bitMaskAncestorRace) > 0 then colorCode = orangeColor else colorCode = normalColor end
			self.tt:AddDoubleLine("|c"..colorCode..RACES.."|r", raceString)

			if 0 ~= holidayCode then
				self.tt:AddLine(" ")
				if bit.band(statusCode, Grail.bitMaskHoliday) > 0 then colorCode = redColor elseif bit.band(statusCode, Grail.bitMaskAncestorHoliday) > 0 then colorCode = orangeColor else colorCode = normalColor end
				self.tt:AddLine("|c"..colorCode..self.s['HOLIDAYS_ONLY'].."|r")
				for letterCode, bitValue in pairs(Grail.holidayToBitMapping) do
					if 0 < bit.band(holidayCode, bitValue) then
						self.tt:AddLine(Grail.holidayMapping[letterCode])
					end
				end
			end

			if Grail.quests[questId]['SP'] then
				self.tt:AddLine(" ")
				self.tt:AddLine(self.s['SP_MESSAGE'])
			end

			if nil ~= Grail.quests[questId]['rep'] then
				local factionStandingFormat = "FACTION_STANDING_LABEL%d"
				if Grail.playerGender == 3 then factionStandingFormat = factionStandingFormat.."_FEMALE" end
				self.tt:AddLine(" ")
				if bit.band(statusCode, Grail.bitMaskReputation) > 0 then colorCode = redColor elseif bit.band(statusCode, Grail.bitMaskAncestorReputation) > 0 then colorCode = orangeColor else colorCode = normalColor end
				self.tt:AddLine("|c"..colorCode..self.s['REPUTATION_REQUIRED'].."|r")
				for reputationIndex, repTable in pairs(Grail.quests[questId]['rep']) do
					-- repTable can have 'min' and/or 'max'
					local reputationString = ""
					local repValue = repTable['min']
					if nil ~= repValue then
						reputationString = format("|cFF00FF00%s|r", GetText(format(factionStandingFormat, self.reputationLevelMapping[repValue])))
					end
					repValue = repTable['max']
					if nil ~= repValue then
						if 0 < strlen(reputationString) then reputationString = reputationString .. " - " end
						reputationString = reputationString .. format("|cFFFF0000%s|r", GetText(format(factionStandingFormat, self.reputationLevelMapping[repValue])))
					end
					self.tt:AddDoubleLine(Grail.reputationMapping[reputationIndex], reputationString)
				end
			end

			if nil ~= Grail.quests[questId]['prof'] then
				self.tt:AddLine(" ")
				if bit.band(statusCode, Grail.bitMaskProfession) > 0 then colorCode = redColor elseif bit.band(statusCode, Grail.bitMaskAncestorProfession) > 0 then colorCode = orangeColor else colorCode = normalColor end
				self.tt:AddLine("|c"..colorCode..TUTORIAL_TITLE38..':'.."|r")
				for code, value in pairs(Grail.quests[questId]['prof']) do
					self.tt:AddDoubleLine(Grail.professionMapping[code], "|cFF00FF00"..value.."|r")
				end
			end

			self:_QuestInfoSection(self.s['BREADCRUMB'], Grail:QuestBreadcrumbs(questId))

--	At the moment the UI will show both invalidated and breadcrumb invalidated ancestors as orange.
			local breadcrumbColorCode
			if bit.band(statusCode, Grail.bitMaskInvalidated) > 0 then
				if Grail:IsInvalidated(questId, true) then	-- still invalid ignoring breadcrumbs
					colorCode = redColor
					breadcrumbColorCode = normalColor
				else
					colorCode = normalColor
					breadcrumbColorCode = redColor
				end
			elseif bit.band(statusCode, Grail.bitMaskAncestorInvalidated) > 0 then
				colorCode = orangeColor
				breadcrumbColorCode = orangeColor
			else
				breadcrumbColorCode = normalColor
				colorCode = normalColor
			end
			self:_QuestInfoSection("|c"..breadcrumbColorCode..self.s['IS_BREADCRUMB'].."|r", Grail:QuestBreadcrumbsFor(questId))			
			self:_QuestInfoSection("|c"..colorCode..self.s['INVALIDATE'].."|r", Grail:QuestInvalidates(questId))

			local lastIndexUsed = 0
			if Grail.DisplayableQuestPrerequisites then
				lastIndexUsed = self:_QuestInfoSection(self.s['PREREQUISITES'], Grail:DisplayableQuestPrerequisites(questId), lastIndexUsed)
			else
				lastIndexUsed = self:_QuestInfoSection(self.s['PREREQUISITES'], Grail:QuestPrerequisites(questId), lastIndexUsed)
			end

			if self.tt:NumLines() > 50 then
				self.tt2:SetOwner(GameTooltip, "ANCHOR_RIGHT")
				self.tt2:ClearLines()
				self.tt = self.tt2
			end

			self:_QuestInfoSection(self.s['OAC'], Grail:QuestOnAcceptCompletes(questId))
			self:_QuestInfoSection(self.s['OCC'], Grail:QuestOnCompletionCompletes(questId))
			self:_QuestInfoTurninSection(self.s['OTC'], Grail:QuestOnTurninCompletes(questId))
			if nil ~= Grail.quests[questId]['AZ'] then
				self.tt:AddLine(" ")
				self.tt:AddDoubleLine(self.s['ENTER_ZONE'], Grail.mapAreaMapping[Grail.quests[questId]['AZ']])
			end

			if nil ~= Grail.quests[questId][6] and 0 < #(Grail.quests[questId][6]) then
				self.tt:AddLine(" ")
				self.tt:AddLine(COMBAT_TEXT_SHOW_REPUTATION_TEXT .. ':')
				local index, value
				for _, repString in pairs(Grail.quests[questId][6]) do
					index = strsub(repString, 1, 3)
					value = tonumber(strsub(repString, 4))
					if value > 0 then
						colorCode = greenColor
					else
						colorCode = redColor
						value = -1 * value
					end
					self.tt:AddDoubleLine(Grail.reputationMapping[index], "|c"..colorCode..value.."|r")
				end
			end

			self:_NPCInfoSection(self.s['WHEN_KILL'], Grail:QuestNPCKills(questId), frame, false)
			self:_NPCInfoSection(TUTORIAL_TITLE1..':', Grail:QuestNPCAccepts(questId), frame, ('I' ~= frame.statusCode))
			self:_NPCInfoSection(TURN_IN_QUEST..':', Grail:QuestNPCTurnins(questId), frame, ('I' == frame.statusCode))
		end,

		_PreparePrerequisiteInfo = function(self, table, lastIndexUsed, doMath)
			if nil == table then return lastIndexUsed end
			local code, numeric
			local index = lastIndexUsed or 0
			for key, value in pairs(table) do
				if doMath then index = index + 1 end
				if "table" == type(value) then
					for key2, value2 in pairs(value) do
						code = ""
						numeric = tonumber(value2)
						if nil == numeric then
							code = strsub(value2, 1, 1)
							numeric = tonumber(strsub(value2, 2))
						end
						if code ~= 'I' and code ~= 'J' and code ~= 'M' and code ~= 'N' and code ~= 'P' then
							self:_GetPrerequisiteInfo(numeric, index)
						end
					end
				else
					code = ""
					numeric = tonumber(value)
					if nil == numeric then
						code = strsub(value, 1, 1)
						numeric = tonumber(strsub(value, 2))
					end
					if code ~= 'I' and code ~= 'J' and code ~= 'M' and code ~= 'N' and code ~= 'P' then
						self:_GetPrerequisiteInfo(numeric, index)
					end
				end
			end
			return index
		end,

		QuestInfoEnter = function(self, frame)
			self:_PopulateTooltipForQuest(frame, self:_BreadcrumbQuestId())
			GameTooltip:Show()
		end,

		_QuestInfoSection = function(self, heading, table, lastUsedIndex)
			if nil == table then return lastUsedIndex end
			self.tt:AddLine(" ")
			self.tt:AddLine(heading)
			local numeric
			local index = lastUsedIndex or 0
			for key, value in pairs(table) do
				index = index + 1
				if "table" == type(value) then
					for key2, value2 in pairs(value) do
						numeric = tonumber(value2)
						if nil == numeric then numeric = tonumber(strsub(value2, 2)) end
						self.tt:AddDoubleLine("("..index..") "..self:_PrettyQuestString({ value2, self:_ClassificationOfQuest(value2) }), numeric)
					end
				else
					numeric = tonumber(value)
					if nil == numeric then numeric = tonumber(strsub(value, 2)) end
					self.tt:AddDoubleLine("("..index..") "..self:_PrettyQuestString({ value, self:_ClassificationOfQuest(value) }), numeric)
				end
			end
			return index
		end,

		_QuestInfoTurninSection = function(self, heading, table)
			if nil == table then return end
			self.tt:AddLine(" ")
			self.tt:AddLine(heading)
			for key, value in pairs(table) do
				if "table" == type(value) and 2 == #value then
					self.tt:AddDoubleLine(Grail:NPCName(value[1]), self:_PrettyQuestString({ value[2], self:_ClassificationOfQuest(value[2]) }).." "..value[2])
				else
					self.tt:AddLine("Internal Error with OTC: ", value)
				end
			end
		end,

		_QuestName = function(self, questId)
			return Grail:QuestName(questId) or "NO NAME"
		end,

		_QuestTypeString = function(self, questId)
			local retval = ""
			local bitValue = Grail:CodeType(questId)
			if bitValue > 0 then
				if bit.band(bitValue, Grail.bitMaskQuestRepeatable) > 0 then retval = retval .. self.s['REPEATABLE'] .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestDaily) > 0 then retval = retval .. DAILY .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestWeekly) > 0 then retval = retval .. CALENDAR_REPEAT_WEEKLY .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestMonthly) > 0 then retval = retval .. CALENDAR_REPEAT_MONTHLY .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestYearly) > 0 then retval = retval .. self.s['YEARLY'] .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestEscort) > 0 then retval = retval .. self.s['ESCORT'] .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestDungeon) > 0 then retval = retval .. CALENDAR_TYPE_DUNGEON .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestRaid) > 0 then retval = retval .. CALENDAR_TYPE_RAID .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestPVP) > 0 then retval = retval .. CALENDAR_TYPE_PVP .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestGroup) > 0 then retval = retval .. CHANNEL_CATEGORY_GROUP .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestHeroic) > 0 then retval = retval .. PLAYER_DIFFICULTY2 .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestScenario) > 0 then retval = retval .. GUILD_CHALLENGE_TYPE4 .. " " end
				if bit.band(bitValue, Grail.bitMaskQuestLegendary) > 0 then retval = retval .. self.s['LEGENDARY'] .. " " end
			end
			return retval
		end,

		--	This records into the npcs table all those NPCs whose tooltips need to be augmented with quest information for the provided mapId.
		_RecordTooltipNPCs = function(self, mapId)
			local questsInMap = self:_ClassifyQuestsInMap(mapId) or {}
			local questId, locations

			for i = 1, #questsInMap do
				questId = questsInMap[i][1]
				locations = Grail:QuestLocations(questId, 'A', false, false, true, mapId, true)
				if nil ~= locations then
					for _, npc in pairs(locations) do
						local xcoord, ycoord, npcName, npcId = npc.x, npc.y, npc.name, npc.id
						if nil ~= xcoord then
							-- record the NPC as needing a tooltip note for the specific quest (it can be a redirect because an actual "NPC" may be the item that starts the quest)
							local shouldProcess, kindsOfNPC = Grail:IsTooltipNPC(npcId)
							if shouldProcess then
								for i = 1, #(kindsOfNPC), 1 do
									local npcIdToUse = npcId
									local shouldAdd = true
									if kindsOfNPC[i][1] == Grail.NPC_TYPE_DROP then
										shouldAdd = self:_DroppedItemMatchesQuest(kindsOfNPC[i][2], questId)
									end
									if kindsOfNPC[i][1] == Grail.NPC_TYPE_BY then npcIdToUse = tonumber(kindsOfNPC[i][2]) end
									if nil == self.npcs[npcIdToUse] then self.npcs[npcIdToUse] = {} end
									if shouldAdd and not tContains(self.npcs[npcIdToUse], questId) then tinsert(self.npcs[npcIdToUse], questId) end
								end
							end
						end
					end
				end
			end
			self.checkedNPCs[mapId] = true
		end,

		--	This walks through the persistent information about groups of waypoints and reinstates
		--	them since our directional arrows we do not have TomTom make persistent.
		_ReinstateDirectionalArrows = function(self)
			if nil == WhollyDatabase.waypointGrouping then return end
			for groupNumber, t in pairs(WhollyDatabase.waypointGrouping) do
				if 0 == #t then
					WhollyDatabase.waypointGrouping[groupNumber] = nil
				else
					local t1 = {}
					local npcType = 'A'
					local codeLen
					for _, code in pairs(t) do
						codeLen = strlen(code)
						tinsert(t1, strsub(code, 1, codeLen - 1))
						npcType = strsub(code, codeLen, codeLen)
					end
					self:_AddDirectionalArrows(t1, npcType, groupNumber)
				end
			end
		end,

		_RemoveAllDirectionalArrows = function(self)
			for code, t in pairs(self.waypoints) do
				WhollyDatabase.waypointGrouping[t.grouping] = nil
			end
			self.waypoints = {}
		end,

		--	This uses the TomTom sense of uid to remove that waypoint and any others that were added
		--	in the same grouping of waypoints.
		_RemoveDirectionalArrows = function(self, uid)
			local foundGrouping = nil
			for code, t in pairs(self.waypoints) do
				if tContains(t.uids, uid) then
					foundGrouping = t.grouping
				end
			end
			if nil ~= foundGrouping then
				for _, code in pairs(WhollyDatabase.waypointGrouping[foundGrouping]) do
					for _, uid in pairs(Wholly.waypoints[code].uids) do
						self.removeWaypointFunction(TomTom, uid)
					end
				end
				WhollyDatabase.waypointGrouping[foundGrouping] = nil
			end
		end,

		ScrollFrame_OnLoad = function(self, frame)
			HybridScrollFrame_OnLoad(frame)
			frame.update = Wholly.ScrollFrame_Update_WithCombatCheck
			HybridScrollFrame_CreateButtons(frame, "com_mithrandir_whollyButtonTemplate")
		end,

		ScrollFrameOne_OnLoad = function(self, frame)
			HybridScrollFrame_OnLoad(frame)
			frame.update = Wholly.ScrollFrameOne_Update
			HybridScrollFrame_CreateButtons(frame, "com_mithrandir_whollyButtonOneTemplate")
		end,

		ScrollFrameTwo_OnLoad = function(self, frame)
			HybridScrollFrame_OnLoad(frame)
			frame.update = Wholly.ScrollFrameTwo_Update
			HybridScrollFrame_CreateButtons(frame, "com_mithrandir_whollyButtonTwoTemplate")
		end,

		ScrollFrame_Update_WithCombatCheck = function(self)
			if not InCombatLockdown() then
				Wholly:ScrollFrame_Update()
			else
				com_mithrandir_whollyFrame:RegisterEvent("PLAYER_REGEN_ENABLED")
			end
		end,

		ScrollFrameOne_Update = function(self)
			self = self or Wholly
--			self:_InitializeLevelTwoData()
			self:ScrollFrameGeneral_Update(self.levelOneData, com_mithrandir_whollyFrameWideScrollOneFrame)
		end,

		ScrollFrameTwo_Update = function(self)
			self = self or Wholly
			self:_InitializeLevelTwoData()
			self:ScrollFrameGeneral_Update(self.levelTwoData, com_mithrandir_whollyFrameWideScrollTwoFrame)
		end,

		SetupScrollFrameButton = function(self, buttonIndex, numButtons, buttons, shownEntries, scrollOffset, item, isHeader, indent, scrollFrame)
			if shownEntries > scrollOffset and buttonIndex <= numButtons then
				local button = buttons[buttonIndex]
				local indentation = indent and "    " or ""
				button.normalText:SetText(indentation .. item.displayName)
				button.tag:SetText(self.cachedMapCounts[item.mapID])
				if WhollyDatabase.showQuestCounts then
					button.tag:Show()
				else
					button.tag:Hide()
				end
				if isHeader then
					if WhollyDatabase.closedHeaders[item.header] then
						button:SetNormalTexture("Interface\\Buttons\\UI-PlusButton-Up")
					else
						button:SetNormalTexture("Interface\\Buttons\\UI-MinusButton-Up")
					end
				else
					button:SetNormalTexture("")
				end
				button.item = item
				local f
				if scrollFrame == com_mithrandir_whollyFrameWideScrollOneFrame then
					f = com_mithrandir_whollyFrameWideScrollOneFrameLogHighlightFrame
				else
					f = com_mithrandir_whollyFrameWideScrollTwoFrameLogHighlightFrame
				end
				if item.selected then
					f:SetParent(button)
					f:SetPoint("TOPLEFT", button, "TOPLEFT", 0, 0)
					f:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 0, 0)
					f:Show()
--					button:LockHighlight()
				else
					if f:GetParent() == button then
						f:Hide()
					end
--					button:UnlockHighlight()
				end
				button:Show()
				buttonIndex = buttonIndex + 1
			end
			return buttonIndex
		end,

		--	This technique uses marching through the data to update the buttons.
		--	This is done because some of our data may be closed, and in any case any of the headings
		--	that are open need to be processed differently.
		ScrollFrameGeneral_Update = function(self, items, frame)
			local numEntries = items and #items or 0
			local shownEntries = 0
			local buttons = frame.buttons
			local numButtons = #buttons
			local scrollOffset = HybridScrollFrame_GetOffset(frame)
			local buttonHeight = buttons[1]:GetHeight()
			local button, itemIndex
			local buttonIndex = 1

			--	Go through the data and put it into the buttons based on where the scrolling is within
			--	the data, and based on what headers are open or closed.
			for i = 1, numEntries do
				if items[i].header then		-- a header
					shownEntries = shownEntries + 1
					buttonIndex = self:SetupScrollFrameButton(buttonIndex, numButtons, buttons, shownEntries, scrollOffset, items[i], true, false, frame)
					if not WhollyDatabase.closedHeaders[items[i].header] then
						for j = 1, #(items[i].children) do
							shownEntries = shownEntries + 1
							buttonIndex = self:SetupScrollFrameButton(buttonIndex, numButtons, buttons, shownEntries, scrollOffset, items[i].children[j], false, true, frame)
						end
					end
				else						-- a normal entry
					shownEntries = shownEntries + 1
					buttonIndex = self:SetupScrollFrameButton(buttonIndex, numButtons, buttons, shownEntries, scrollOffset, items[i], false, false, frame)
				end
				
			end

			--	Now any remaining buttons in the UI should be hidden
			for i = buttonIndex, numButtons do
				buttons[i]:Hide()
			end

			--	How have the scroll frame update itself
			HybridScrollFrame_Update(frame, shownEntries * buttonHeight, numButtons * buttonHeight)
		end,

		ScrollFrame_Update = function(self)
			self = self or Wholly
			self:_FilterPanelQuests()
			local questsInMap = self.filteredPanelQuests
			local numEntries = #questsInMap
			local buttons = com_mithrandir_whollyFrameScrollFrame.buttons
			local numButtons = #buttons
			local scrollOffset = HybridScrollFrame_GetOffset(com_mithrandir_whollyFrameScrollFrame)
			local buttonHeight = buttons[1]:GetHeight();
			local button, questIndex, questId, questLevelString, requiredLevelString, colorCode, questLevel, filterCode, repeatableCompletedString
			local shouldShowTag

			table.sort(questsInMap, Wholly.SortingFunction)
			for i = 1, numButtons do
				button = buttons[i]
				questIndex = i + scrollOffset
				if questIndex <= numEntries then
					questId = questsInMap[questIndex][1]
					filterCode = questsInMap[questIndex][2]
					button.normalText:SetText(self:_PrettyQuestString(questsInMap[questIndex]))
					shouldShowTag = false
					if 'I' == filterCode and WhollyDatabase.showsInLogQuestStatus and WhollyDatabase.showsQuestsInLog then
						local questStatus = Grail:StatusCode(questId)
						local statusText = nil
						shouldShowTag = true
						if bit.band(questStatus, Grail.bitMaskInLogComplete) > 0 then statusText = COMPLETE
						elseif bit.band(questStatus, Grail.bitMaskInLogFailed) > 0 then statusText = FAILED
						else shouldShowTag = false
						end
						if nil ~= statusText then
							button.tag:SetText("(" .. statusText .. ")")
						end
					end
					if shouldShowTag then button.tag:Show() else button.tag:Hide() end
					button.questId = questId
					button.statusCode = filterCode
					button:Show()
				else
					button:Hide()
				end
			end
			HybridScrollFrame_Update(com_mithrandir_whollyFrameScrollFrame, numEntries * buttonHeight, numButtons * buttonHeight)
		end,

		ScrollOneClick = function(self, button)
			if button.item.header then
				local which = button.item.header
				if WhollyDatabase.closedHeaders[which] then
					WhollyDatabase.closedHeaders[which] = nil
				else
					WhollyDatabase.closedHeaders[which] = true
				end
				self:ScrollFrameOne_Update()
			else
				if self.levelOneCurrent ~= button.item then
					self.currentPanelMapID = nil
					self:_SetLevelTwoCurrent(nil)
					self:_ForcePanelMapArea(true)
				end
				self:_SetLevelOneCurrent(button.item)
				self:ScrollFrameOne_Update()
				self:ScrollFrameTwo_Update()
			end
		end,

		ScrollTwoClick = function(self, button)
			self:_SetLevelTwoCurrent(button.item)
			if button.item.f then
				button.item.f()
			else
				self.currentPanelMapID = button.item.mapID
				self:_ForcePanelMapArea(true)
			end
			self:ScrollFrameTwo_Update()	-- to update selection
		end,

		SearchEntered = function(self)
			local searchText = com_mithrandir_whollySearchEditBox:GetText()
			com_mithrandir_whollySearchEditBox:SetText("")
			com_mithrandir_whollySearchFrame:Hide()

			-- Remove leading and trailing whitespace
			searchText = searchText:gsub("^%s*(.-)%s*$", "%1")

			if searchText and "" ~= searchText then
				if nil == WhollyDatabase.searches then WhollyDatabase.searches = {} end
				tinsert(WhollyDatabase.searches, searchText)
				if #(WhollyDatabase.searches) > self.maximumSearchHistory then
					tremove(WhollyDatabase.searches, 1)
				end
				self:SearchForQuestNamesMatching(searchText)
				self.currentPanelMapID = 0
				self.justAddedSearch = true
				self:_ForcePanelMapArea(true)
				self:ScrollFrameTwo_Update()
			end
		end,

		SearchForAllQuests = function(self)
			Grail.indexedQuests[0] = {}
			Grail.mapAreaMapping[0] = SEARCH .. ' ' .. Wholly.s['SEARCH_ALL_QUESTS']
			for q, v in pairs(Grail.quests) do
				Grail:AddQuestToMapArea(q, 0)
			end
		end,

		SearchForQuestNamesMatching = function(self, searchTerm)
			-- the searchTerm is broken up by spaces which are considered AND conditions
			local terms = { strsplit(" ", searchTerm) }
			local results = {}
			local started = false
			for i = 1, #terms do
				if terms[i] ~= "" then
					if not started then
						for q, v in pairs(Grail.quests) do
							if strfind(v[0], terms[i]) then tinsert(results, q) end
						end
						started = true
					else
						local newResults = {}
						for _, q in pairs(results) do
							if strfind(Grail.quests[q][0], terms[i]) then tinsert(newResults, q) end
						end
						results = newResults
					end
				end
			end
			-- clear the mapArea 0 because that is what we use for computed results
			Grail.indexedQuests[0] = {}
			Grail.mapAreaMapping[0] = SEARCH .. ': ' .. searchTerm
			for _, q in pairs(results) do
				Grail:AddQuestToMapArea(q, 0)
			end
		end,

		SetCurrentZone = function(self, frame)		-- called by pressing the Zone button in the UI
			self:UpdateQuestCaches(false, false, true)
			self:ZoneButtonEnter(frame)	-- need to update the tooltip which is showing
		end,

		_SetLevelOneCurrent = function(self, newValue)
			if self.levelOneCurrent ~= newValue then
				if self.levelOneCurrent ~= nil then
					self.levelOneCurrent.selected = nil
				end
				self.levelOneCurrent = newValue
				if newValue ~= nil then
					newValue.selected = true
				end
			end
		end,

		_SetLevelTwoCurrent = function(self, newValue)
			if self.levelTwoCurrent ~= newValue then
				if self.levelTwoCurrent ~= nil then
					self.levelTwoCurrent.selected = nil
				end
				self.levelTwoCurrent = newValue
				if newValue ~= nil then
					newValue.selected = true
				end
			end
		end,

		ShowBreadcrumbInfo = function(self)
			local questId = self:_BreadcrumbQuestId()
			local breadcrumbs = Grail:AvailableBreadcrumbs(questId)
			com_mithrandir_whollyBreadcrumbFrame:Hide()
			com_mithrandir_whollyQuestInfoFrameText:SetText(questId)
			self:UpdateBuggedText(questId)
			if nil ~= breadcrumbs then
				if 1 == #breadcrumbs then com_mithrandir_whollyBreadcrumbFrameMessage:SetText(self.s['SINGLE_BREADCRUMB_FORMAT'])
				else com_mithrandir_whollyBreadcrumbFrameMessage:SetText(format(self.s['MULTIPLE_BREADCRUMB_FORMAT'], #breadcrumbs))
				end
				com_mithrandir_whollyBreadcrumbFrame:Show()
			end
		end,

		ShowPin = function(self, questTable)
			local codeMapping = { ['G'] = 1, ['W'] = 2, ['D'] = 3, ['R'] = 4, ['H'] = 5, ['Y'] = 6, ['P'] = 7, ['L'] = 8, ['U'] = 9 }
			local id = questTable[1]
			local locations = Grail:QuestLocations(id, 'A', false, false, true, self.currentPinMapID, true, self.currentPinMapLevel)
			if nil ~= locations then
				for _, npc in pairs(locations) do
					local xcoord, ycoord, npcName, npcId = npc.x, npc.y, npc.name, npc.id
					if nil ~= xcoord then

						if not self.checkingNPCTechniqueNew then
							-- record the NPC as needing a tooltip note for the specific quest (it can be a redirect because an actual "NPC" may be the item that starts the quest)
							local shouldProcess, kindsOfNPC = Grail:IsTooltipNPC(npcId)
							if shouldProcess then
								for i = 1, #(kindsOfNPC), 1 do
									local npcIdToUse = npcId
									local shouldAdd = true
									if kindsOfNPC[i][1] == Grail.NPC_TYPE_DROP then
										shouldAdd = self:_DroppedItemMatchesQuest(kindsOfNPC[i][2], id)
									end
									if kindsOfNPC[i][1] == Grail.NPC_TYPE_BY then npcIdToUse = tonumber(kindsOfNPC[i][2]) end
									if nil == self.npcs[npcIdToUse] then self.npcs[npcIdToUse] = {} end
									if shouldAdd and not tContains(self.npcs[npcIdToUse], id) then tinsert(self.npcs[npcIdToUse], id) end
								end
							end
						end

						local pin = self:_GetPin(npcId)
						local code = questTable[2]
						if 'D' == code and Grail:IsRepeatable(id) then code = 'R' end
						local codeValue = codeMapping[code]
						local pinValue = codeMapping[pin.texType]
						if codeValue < pinValue then
							pin:SetType(code)
						end						
						pin:ClearAllPoints()
						pin.questId = id
						pin:SetFrameLevel(WorldMapPOIFrame:GetFrameLevel() - 1)
						pin:SetPoint("CENTER", WorldMapDetailFrame, "TOPLEFT", xcoord/100*WorldMapDetailFrame:GetWidth(), -ycoord/100*WorldMapDetailFrame:GetHeight())
						pin:Show()

						self.regPins[id..":"..npcId] = pin
					end
				end
			end
		end,

		ShowTooltip = function(self, pin)
			local listedQuests = {}
			self.tooltip:SetOwner(pin, "ANCHOR_RIGHT")
			self.tooltip:ClearLines()
	
			-- find all quests in range of hover
			local mx, my = self:_GetMousePosition()
			local npcList = {}
			local npcNames = {}

			local mapId = self.currentPinMapID
			local mapLevel = self.currentPinMapLevel
			local questsInMap = self.filteredPinQuests
			local questId
			for i = 1, #questsInMap do
				questId = questsInMap[i][1]
				local locations = Grail:QuestLocations(questId, 'A', false, false, true, mapId, true, mapLevel)
				if nil ~= locations then
					for _, npc in pairs(locations) do
						if nil ~= npc.x then
							local dist = self:_Distance(mx, my, npc.x/100, npc.y/100)
							if dist <= 0.02 then
								if not npcList[npc.id] then
									npcList[npc.id] = {}
									local nameToUse = npc.name
									if npc.dropName then
										nameToUse = nameToUse .. " (" .. npc.dropName .. ')'
									end
									npcNames[npc.id] = self:_PrettyNPCString(nameToUse, npc.kill, npc.realArea)
								end
								table.insert(npcList[npc.id], questsInMap[i])
							end
						end
					end
				end
			end

			local first = true
			for npc, questList in pairs(npcList) do
				if not first then
					self.tooltip:AddLine(" ")
				else
					first = false
				end
				for _, qt in ipairs(questList) do
					local leftStr = self:_PrettyQuestString(qt)
					local q = qt[1]
					local rightStr = self:_QuestTypeString(q):gsub("^%s*(.-)%s*$", "%1") -- remove the leading and trailing whitespace
					if strlen(rightStr) > 0 then rightStr = format("|c%s%s|r", self.color[qt[2]], rightStr) end

					-- check if already printed - this is for spam quests like the human starting area that haven't been labeled correctly
					if not questName or not listedQuests[questName] then
						self.tooltip:AddDoubleLine(leftStr, rightStr)
						self.tooltip:SetLastFont(self.tooltip.large)
						self.tooltip:SetLastFont(self.tooltip.small, true)
						if questName then listedQuests[questName] = true end
					end
				end
				self.tooltip:AddLine(npcNames[npc], 1, 1, 1, 1)
				self.tooltip:SetLastFont(self.tooltip.small)
			end
	
			self.tooltip:Show();
		end,

		SlashCommand = function(self, frame, msg)
			self:ToggleUI()
		end,

		Sort = function(self, frame)
			-- This is supposed to cycle through the supported sorting techniques and make the contents of the panel
			-- show the quests based on those techniques.
			-- 1 Quest alphabetical
			-- 2 Quest level (and then alphabetical)
			-- 3 Quest level, then type, then alphabetical
			-- 4 Quest type (and then alphabetical)
			-- 5 Quest type, then level, then alphabetical
			WhollyDatabase.currentSortingMode = WhollyDatabase.currentSortingMode + 1
			if (WhollyDatabase.currentSortingMode > 5) then WhollyDatabase.currentSortingMode = 1 end
			self:ScrollFrame_Update_WithCombatCheck()
			self:SortButtonEnter(frame)	-- to update the tooltip with the new sorting info
		end,

		SortButtonEnter = function(self, frame)
			local sortModes = {
				[1] = COMPACT_UNIT_FRAME_PROFILE_SORTBY_ALPHABETICAL,
				[2] = LEVEL..", "..COMPACT_UNIT_FRAME_PROFILE_SORTBY_ALPHABETICAL,
				[3] = LEVEL..", "..TYPE..", "..COMPACT_UNIT_FRAME_PROFILE_SORTBY_ALPHABETICAL,
				[4] = TYPE..", "..COMPACT_UNIT_FRAME_PROFILE_SORTBY_ALPHABETICAL,
				[5] = TYPE..", "..LEVEL..", "..COMPACT_UNIT_FRAME_PROFILE_SORTBY_ALPHABETICAL,
				}
			GameTooltip:ClearLines()
			GameTooltip:SetOwner(frame, "ANCHOR_RIGHT")
			GameTooltip:AddLine(sortModes[WhollyDatabase.currentSortingMode])
			GameTooltip:Show()
			GameTooltip:ClearAllPoints()
		end,

		SortingFunction = function(a, b)
			local retval = false
			if 1 == WhollyDatabase.currentSortingMode then
				retval = Wholly:_QuestName(a[1]) < Wholly:_QuestName(b[1])
			elseif 2 == WhollyDatabase.currentSortingMode then
				local aLevel, bLevel = Grail:QuestLevel(a[1]) or 1, Grail:QuestLevel(b[1]) or 1
				if aLevel == bLevel then
					retval = Wholly:_QuestName(a[1]) < Wholly:_QuestName(b[1])
				else
					retval = aLevel < bLevel
				end
			elseif 3 == WhollyDatabase.currentSortingMode then
				local aLevel, bLevel = Grail:QuestLevel(a[1]) or 1, Grail:QuestLevel(b[1]) or 1
				if aLevel == bLevel then
					local aCode, bCode = a[2], b[2]
					if aCode == bCode then
						retval = Wholly:_QuestName(a[1]) < Wholly:_QuestName(b[1])
					else
						retval = aCode < bCode
					end
				else
					retval = aLevel < bLevel
				end
			elseif 4 == WhollyDatabase.currentSortingMode then
				local aCode, bCode = a[2], b[2]
				if aCode == bCode then
					retval = Wholly:_QuestName(a[1]) < Wholly:_QuestName(b[1])
				else
					retval = aCode < bCode
				end
			elseif 5 == WhollyDatabase.currentSortingMode then
				local aCode, bCode = a[2], b[2]
				if aCode == bCode then
					local aLevel, bLevel = Grail:QuestLevel(a[1]) or 1, Grail:QuestLevel(b[1]) or 1
					if aLevel == bLevel then
						retval = Wholly:_QuestName(a[1]) < Wholly:_QuestName(b[1])
					else
						retval = aLevel < bLevel
					end
				else
					retval = aCode < bCode
				end
			end
			return retval
		end,

		ToggleCurrentFrame = function(self)
			local isShowing = self.currentFrame:IsShown()
			local x, y
			if isShowing then
				self.currentFrame:Hide()	-- Hide the current frame before we manipulate
			end
			if com_mithrandir_whollyFrame == self.currentFrame then
				self.currentFrame = com_mithrandir_whollyFrameWide
				x, y = 348, -75
			else
				self.currentFrame = com_mithrandir_whollyFrame
				x, y = 19, -75
			end
			com_mithrandir_whollyFrameScrollFrame:ClearAllPoints()
			com_mithrandir_whollyFrameScrollFrame:SetParent(self.currentFrame)
			com_mithrandir_whollyFrameScrollFrame:SetPoint("TOPLEFT", self.currentFrame, "TOPLEFT", x, y)
			if isShowing then
				self.currentFrame:Show()
			end
			com_mithrandir_whollySearchFrame:ClearAllPoints()
			com_mithrandir_whollySearchFrame:SetParent(self.currentFrame)
			com_mithrandir_whollySearchFrame:SetPoint("BOTTOMLEFT", self.currentFrame, "TOPLEFT", 64, -14)
		end,

		ToggleLightHeaded = function(self)
			local desiredQuestId = self.clickingButton.questId
			if LightHeadedFrame:IsVisible() and LightHeadedFrameSub.qid == desiredQuestId then LightHeadedFrame:Hide() return end
			LightHeadedFrame:ClearAllPoints()
			LightHeadedFrame:SetParent(self.currentFrame)
			-- default to attaching on the right side
			local lhSide, whollySide, x, y = "LEFT", "RIGHT", -39, 31
			LightHeadedFrame:SetPoint(lhSide, self.currentFrame, whollySide, x, y)
			LightHeadedFrame:Show()
			LightHeaded:UpdateFrame(desiredQuestId, 1)
		end,

		ToggleUI = function(self)
			if not InCombatLockdown() then
				if not self.currentFrame then print(format(self.s['REQUIRES_FORMAT'], requiredGrailVersion)) return end
				if self.currentFrame:IsShown() then
					self.currentFrame:Hide()
				else
					self.currentFrame:Show()
				end
			end
		end,

		---
		--	Sets up the event monitoring to handle those associated with displaying breadcrumb information.
		UpdateBreadcrumb = function(self)
			if WhollyDatabase.displaysBreadcrumbs then
				self.notificationFrame:RegisterEvent("QUEST_DETAIL")
				if QuestFrame:IsVisible() then
					self:ShowBreadcrumbInfo()
				end
			else
				self.notificationFrame:UnregisterEvent("QUEST_DETAIL")
				com_mithrandir_whollyBreadcrumbFrame:Hide()
			end
		end,

		UpdateBuggedText = function(self, questId)
			local bugged = Grail:IsBugged(questId)
			if bugged then
				com_mithrandir_whollyQuestInfoBuggedFrameText:SetText(self.s['BUGGED'])
			else
				com_mithrandir_whollyQuestInfoBuggedFrameText:SetText("")
			end
		end,

		UpdateCoordinateSystem = function(self)
			if WhollyDatabase.enablesPlayerCoordinates then
				self.notificationFrame:SetScript("OnUpdate", function(frame, ...) self:_OnUpdate(frame, ...) end)
			else
				self.notificationFrame:SetScript("OnUpdate", nil)
				if nil ~= self.coordinates then
					self.coordinates.text = ""
				end
				self.previousX = 0
			end
		end,

		_UpdatePins = function(self, forceUpdate)

			--	Set the current mapId to be something it cannot be normally to force an update
			if forceUpdate then
				self.currentPinMapID = -123
			end

			--	Only do work if the world map is visible
			if WorldMapFrame:IsVisible() then

				local currentMapId = GetCurrentMapAreaID()
				local pinsShouldBeReclassified = (self.currentPinMapID ~= currentMapId)

				--	If we are not displaying pins or if anything has changed since we last displayed
				--	pins, we need to hide (remove from the map) all the current pins.
				if not WhollyDatabase.displaysMapPins or pinsShouldBeReclassified or self.pinsNeedFiltering then
					self:_HideAllPins()
				end

				--	If we are displaying pins and something has changed since we last displayed
				--	pins, we need to display all the current pins.
				if WhollyDatabase.displaysMapPins and (pinsShouldBeReclassified or self.pinsNeedFiltering or self.pinsDisplayedLast ~= WhollyDatabase.displaysMapPins) then
					self.currentPinMapID = currentMapId
					self.currentPinMapLevel = GetCurrentMapDungeonLevel() or 0
					if pinsShouldBeReclassified then
						self.cachedPinQuests = self:_ClassifyQuestsInMap(self.currentPinMapID) or {}
					end
					self:_FilterPinQuests()
					self.pinsNeedFiltering = false
					local questsInMap = self.filteredPinQuests
					for i = 1, #questsInMap do
						self:ShowPin(questsInMap[i])
					end
				else
					self.mapCountLine = ""		-- do not display a tooltip for pins we are not showing
				end

				self.pinsDisplayedLast = WhollyDatabase.displaysMapPins

			end

		end,

--		UpdatePins = function(self)
--			self:_HideAllPins()
--			if WhollyDatabase.displaysMapPins then
--				self:_FilterPinQuests()
--				local questsInMap = self.filteredPinQuests
--				for i = 1, #questsInMap do
--					self:ShowPin(questsInMap[i])
--				end
--			else
--				if 0 == #(self.filteredPinQuests) then self:_FilterPinQuests() end
--				self.mapCountLine = self:_PrettyQuestCountString(self.cachedPinQuests, #(self.filteredPinQuests), true)
--			end
--		end,

		UpdateQuestCaches = function(self, forceUpdate, setPinMap, setPanelMap)
			if not Grail:IsPrimed() then return end
			local mapArea = GetCurrentMapAreaID()
--			local dungeonLevel = GetCurrentMapDungeonLevel() or 0
--			if mapArea ~= self.currentPinMapID or dungeonLevel ~= self.currentPinMapLevel or forceUpdate then
--				if setPinMap then self.currentPinMapID = mapArea self.currentPinMapLevel = dungeonLevel end
--				self.cachedPinQuests = self:_ClassifyQuestsInMap(self.currentPinMapID) or {}
--				self:UpdatePins()
--			end
			if mapArea ~= self.currentPanelMapID or forceUpdate then
				if setPanelMap then self.currentPanelMapID = mapArea end
				self:_ForcePanelMapArea(not setPanelMap)
			end
		end,

		ZoneButtonEnter = function(self, frame)
			GameTooltip:ClearLines()
			GameTooltip:SetOwner(frame, "ANCHOR_RIGHT")
			GameTooltip:AddLine(Wholly.panelCountLine)
			GameTooltip:Show()
			GameTooltip:ClearAllPoints()
		end,

		}

	local locale = GetLocale()
	local C = Wholly.configuration
	if "deDE" == locale then
		C[1][0]="Hauptquests"
		C[2][0]="|cFF00FF00Abgeschlossen|r"
		C[3][0]="|cFFFF0000Benötigt Vorraussetzungen|r"
		C[4][0]="|cFF996600Unerfüllbar|r"
		C[6][0]="|cFF0099CCWiederholbar|r"
		C[8][0]="|cFFFF00FFIm Log|r"
		C[9][0]="|cFF666666Niedrigstufig|r"
		C[10][0]="Hochstufig"
		C[11][0]="Aussehen der Quests im Quest-Log"
		C[12][0]="Queststufe voranstellen"
		C[13][0]="Erforderliche Stufe anhängen"
		C[14][0]="Zeige ob wiederholbare Quests bereits fertiggestellt wurden"
		C[15][0]="Zeige Status der Quests im Log"
		C[17][0]="Zeige Karten-Symbole für Questgeber an"
		C[18][0]="Zeige Button auf Weltkarte"
		C[19][0]="Zeige Dungeon-Quests in Umgebungskarte"
		C[20][0]="Weltkarte aktualisieren, wenn Zone wechselt"
		C[21][0]="Sonstiges"
		C[22][0]="Quest-Log aktualisieren, wenn Zone wechselt"
		C[23][0]="Zeige detaillierte Quest-Informationen im Questfenster"
		C[24][0]="Zeige Feiertagsquests immer"
		C[25][0]="Zeige nur 'Meister der Lehren' Quests"
		C[26][0]="Anzeige der Spieler Koordinaten"
		C[27][0]="Zeige Farben je nach Erfolgs-Vervollständigung"
		C[28][0]="Fehlerhafte, wahrscheinlich unerfüllbare Quests"
		C[29][0]="Wide Wholly Quest Panel"
		C[30][0]="Show"
		C[31][0]='Show quest counts'
		C[32][0]='Live quest count updates'
		C[33][0]="Daten laden"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Töte um [%s] zu starten",
			['DROP_TO_START_FORMAT'] = "Dropt %s um [%s] zu starten",
			['REQUIRES_FORMAT'] = "Wholly benötigt Grail-Version %s oder neuer",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Eskorte",
			['BREADCRUMB'] = "Brotkrumen Quests:",
			['IS_BREADCRUMB'] = "Is breadcrumb quest for:",
			['PREREQUISITES'] = "Quests, die Vorraussetzung sind:",
			['OTHER'] = "Andere",
			['SINGLE_BREADCRUMB_FORMAT'] = "Brotkrumen-Quest verfügbar",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Brotkrumen-Quests verfügbar",
			['WORLD_EVENTS'] = "Weltereignisse",
			['REPUTATION_REQUIRED'] = "Ruf erforderlich:",
			['REPEATABLE'] = "Wiederholbar",
			['YEARLY'] = "Jährlich",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Grail kennt diese Quest nicht|r",
			['QUEST_ID'] = "Quest ID: ",
			['REQUIRED_LEVEL'] = "Benötigte Stufe",
			['MAXIMUM_LEVEL_NONE'] = "Keine",
			['QUEST_TYPE_NORMAL'] = "Normal",
			['MAPAREA_NONE'] = "Keine",
			['LOREMASTER_AREA'] = "'Meister der Lehren' Bereich",
			['FACTION_BOTH'] = "Beide",
			['CLASS_NONE'] = "Keine",
			['CLASS_ANY'] = "Jede",
			['GENDER_NONE'] = "Keins",
			['GENDER_BOTH'] = "Beide",
			['GENDER'] = "Geschlecht",
			['RACE_NONE'] = "Keine",
			['RACE_ANY'] = "Jede",
			['HOLIDAYS_ONLY'] = "Verfügbar nur an Feiertagen:",
			['SP_MESSAGE'] = "Spezial-Quest taucht niemals in Blizzards Quest-Log auf",
			['INVALIDATE'] = "Ungültig durch Quests:",
			['OAC'] = "Bei Annahme fertiggestellte Quests:",
			['OCC'] = "Bei Erfüllung der Voraussetzungen fertiggestellte Quests:",
			['OTC'] = "Beim Abgeben fertiggestellte Quests:",
			['ENTER_ZONE'] = "Annahme, wenn Kartenbereich erreicht wird",
			['WHEN_KILL'] = "Annahme beim Töten:",
			['SEARCH_NEW'] = "Neue Suche",
			['SEARCH_CLEAR'] = "Suche löschen",
			['SEARCH_ALL_QUESTS'] = "All quests",
			['NEAR'] = "Naher NPC",
			['FIRST_PREREQUISITE'] = "Erster in einer Questreihe:",
			['BUGGED'] = "|cffff0000*** FEHLERHAFT ***|r",
			['IN_LOG'] = "Im Log",
			['TURNED_IN'] = "Abgegeben",
			['EVER_COMPLETED'] = "Wurde bereits einmal fertiggestellt",
			['ITEM'] = "Gegenstand",
			['ITEM_LACK'] = "Gegenstand fehlt",
			['ABANDONED'] = "Abgebrochen",
			['NEVER_ABANDONED'] = "Niemals abgebrochen",
			['ACCEPTED'] = "Angenommen",
			['LEGENDARY'] = "Legendär",
			}
	elseif "esES" == locale or "esMX" == locale then
		C[1][0]="Búsqueda Base"
		C[2][0]="|cFF00FF00Completada|r"
		C[3][0]="|cFFFF0000Necesita prerequisitos|r"
		C[4][0]="|cFF996600No obtenible|r"
		C[6][0]="|cFF0099CCRepetible|r"
		C[8][0]="|cFFFF00FFEn el Registro|r"
		C[9][0]="|cFF666666Nivel-bajo|r"
		C[10][0]="Nivel Alto"
		C[11][0]="Apariencia del Título de Búsqueda"
		C[12][0]="Anteponer nivel de la búsqueda"
		C[13][0]="Añadir nivel requerido"
		C[14][0]="Mostrar si búsquedas repetibles han sido completadas"
		C[15][0]="Mostrar estatus de las búsquedas en el registro"
		C[17][0]="Demostrar flechas para dadores de búsquedas"
		C[18][0]="Demostrar botón en el mapa mundial"
		C[19][0]="Demostrar busqúedas de mazmorras en el mapa exterior"
		C[20][0]="Se actualiza el mapa mundial cuando se cambia de zona"
		C[21][0]="Otra"
		C[22][0]="Actualizar registro de búsquedas cuando se cambia de zona"
		C[23][0]="Demostrar la información del sendero de la búsqueda en la Cuadra de la Búsqueda"
		C[24][0]="Siempre demostrar búsquedas de feriados"
		C[25][0]="Solo mostrar búsquedas del Maestro del Saber "
		C[26][0]="Habilitar coordenadas del jugador"
		C[27][0]="Mostrar colores de competación de logros"
		C[28][0]="Búsquedas con errores se consideran no obtenibles"
		C[29][0]="Wide Wholly Quest Panel"
		C[30][0]="Show"
		C[31][0]='Show quest counts'
		C[32][0]='Live quest count updates'
		C[33][0]="Cargar Data"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Matar para iniciar [%s]",
			['DROP_TO_START_FORMAT'] = "Deja caer %s para iniciar [%s]",
			['REQUIRES_FORMAT'] = "Wholly requiere la versión %s o mas reciente de Grail",
			['MUST_KILL_PIN_FORMAT'] = "%s [Matar]",
			['ESCORT'] = "Acompañar",
			['BREADCRUMB'] = "Búsquedas de senderos migas de pan:",
			['IS_BREADCRUMB'] = "Is breadcrumb quest for:",
			['PREREQUISITES'] = "Búsquedas prerequisitas:",
			['OTHER'] = "Otro",
			['SINGLE_BREADCRUMB_FORMAT'] = "Búsqueda de senderos migas de pan disponibles",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Búsquedas de sendero de migas de pan disponibles",
			['WORLD_EVENTS'] = "Eventos Mundiales",
			['REPUTATION_REQUIRED'] = "Reputación Requerida:",
			['REPEATABLE'] = "Repetible",
			['YEARLY'] = "Anualmente",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Grail no tiene esta búsqueda|r",
			['QUEST_ID'] = "ID de Búsqueda: ",
			['REQUIRED_LEVEL'] = "Nivel Requerido",
			['MAXIMUM_LEVEL_NONE'] = "Ninguno",
			['QUEST_TYPE_NORMAL'] = "Normal",
			['MAPAREA_NONE'] = "Ninguno",
			['LOREMASTER_AREA'] = "Area del Maestro del Saber",
			['FACTION_BOTH'] = "Ambos",
			['CLASS_NONE'] = "Ninguno",
			['CLASS_ANY'] = "Todos",
			['GENDER_NONE'] = "Ninguno",
			['GENDER_BOTH'] = "Ambos",
			['GENDER'] = "Sexo",
			['RACE_NONE'] = "Ninguna",
			['RACE_ANY'] = "Cualquiera",
			['HOLIDAYS_ONLY'] = "Solo disponible durante feriados:",
			['SP_MESSAGE'] = "Búsqueda especial nunca entra al registro de búsquedas de Blizzard",
			['INVALIDATE'] = "Hecho invalido por búsquedas:",
			['OAC'] = "Al aceptar completa búsqueda:",
			['OCC'] = "Al cumplir los requisitos completa búsquedas:",
			['OTC'] = "Al entregar completa búsquedas:",
			['ENTER_ZONE'] = "Aceptado al entrar area del mapa",
			['WHEN_KILL'] = "Aceptado al matar:",
			['SEARCH_NEW'] = "Nueva",
			['SEARCH_CLEAR'] = "Aclarar",
			['SEARCH_ALL_QUESTS'] = "Todas búsquedas",
			['NEAR'] = "Cerca",
			['FIRST_PREREQUISITE'] = "Primero en la Cadena de Prerequisitos:",
			['BUGGED'] = "|cffff0000*** ERROR ***|r",
			['IN_LOG'] = "En el Registro",
			['TURNED_IN'] = "Entregado",
			['EVER_COMPLETED'] = "Ya completado",
			['ITEM'] = "Articulo",
			['ITEM_LACK'] = "Articulo falta",
			['ABANDONED'] = "Abandonado",
			['NEVER_ABANDONED'] = "Nunca Abandonado",
			['ACCEPTED'] = "Aceptado",
			['LEGENDARY'] = "Legendary",
			}
	elseif "frFR" == locale then
		C[1][0]="Quêtes de base"
		C[2][0]="|cFF00FF00Rendues|r"
		C[3][0]="|cFFFF0000Prérequis nécessaires|r"
		C[4][0]="|cFF996600Impossible à obtenir|r"
		C[6][0]="|cFF0099CCRépétables|r"
		C[8][0]="|cFFFF00FFDans le journal|r"
		C[9][0]="|cFF666666Bas niveau|r"
		C[10][0]="Haut niveau"
		C[11][0]="Apparence du titre des quêtes"
		C[12][0]="Ajouter le niveau de la quête avant son nom"
		C[13][0]="Ajouter le niveau minimum requis après le nom de la quête"
		C[14][0]="Afficher si les quêtes répétables ont déjà terminé auparavant"
		C[15][0]="Afficher l'état des quêtes dans le journal"
		C[17][0]="Afficher les icônes ! des donneurs de quêtes sur la carte"
		C[18][0]="Afficher le bouton sur la carte du monde"
		C[19][0]="Afficher les quêtes de donjons sur la carte du monde"
		C[20][0]="Mise à jour de la carte du monde lors d'un changement de zone"
		C[21][0]="Autres"
		C[22][0]="Mise à jour du journal de quêtes lors d'un changement de zone"
		C[23][0]="Afficher les informations d'une suite de quêtes dans le journal de quêtes"
		C[24][0]="Toujours afficher les quêtes des événements mondiaux"
		C[25][0]="Afficher uniquement les quêtes comptant pour le haut fait de \"Maître des traditions\""
		C[26][0]="Activer les coordonnées du joueur"
		C[27][0]="Afficher les couleurs de l'avancement de la progression des objectifs de hauts faits"
		C[28][0]="Quêtes boguées considérées impossibles à obtenir"
		C[29][0]="Wide Wholly Quest Panel"
		C[30][0]="Show"
		C[31][0]='Show quest counts'
		C[32][0]='Live quest count updates'
		C[33][0]="Chargement des données"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Tuer pour commencer [%s]",
			['DROP_TO_START_FORMAT'] = "Ramasser %s (butin) pour commencer [%s]",
			['REQUIRES_FORMAT'] = "Wholly nécessite Grail version %s ou ultérieure",
			['MUST_KILL_PIN_FORMAT'] = "%s [Tuer]",
			['ESCORT'] = "Escorte",
			['BREADCRUMB'] = "Quêtes précédentes (suite de quêtes) :",
			['IS_BREADCRUMB'] = "Is breadcrumb quest for:",
			['PREREQUISITES'] = "Quêtes préalables :",
			['OTHER'] = "Autres",
			['SINGLE_BREADCRUMB_FORMAT'] = "Quête précédente disponible (suite de quêtes)",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d quêtes précédentes disponibles (suite de quêtes)",
			['WORLD_EVENTS'] = "Événements mondiaux",
			['REPUTATION_REQUIRED'] = "Réputation nécessaire :",
			['REPEATABLE'] = "Répétable",
			['YEARLY'] = "Annuelle",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Grail n'a pas cette quête dans sa base de données|r",
			['QUEST_ID'] = "ID de quête : ",
			['REQUIRED_LEVEL'] = "Niveau requis",
			['MAXIMUM_LEVEL_NONE'] = "Aucun",
			['QUEST_TYPE_NORMAL'] = "Normal",
			['MAPAREA_NONE'] = "Aucun",
			['LOREMASTER_AREA'] = "Zone de maître des traditions",
			['FACTION_BOTH'] = "Les deux",
			['CLASS_NONE'] = "Aucune",
			['CLASS_ANY'] = "Toutes",
			['GENDER_NONE'] = "Aucun",
			['GENDER_BOTH'] = "Les deux",
			['GENDER'] = "Sexe",
			['RACE_NONE'] = "Aucune",
			['RACE_ANY'] = "Toutes",
			['HOLIDAYS_ONLY'] = "Disponible uniquement pendant un événement mondial :",
			['SP_MESSAGE'] = "Certaines quêtes spéciales ne sont jamais afficher dans le journal de quêtes de Blizzard",
			['INVALIDATE'] = "Invalidée de par les quêtes :",
			['OAC'] = "Quêtes complétées par acceptation :",
			['OCC'] = "Quêtes complétées par objectifs atteints :",
			['OTC'] = "Quêtes complétées de par être rendues :",
			['ENTER_ZONE'] = "Acceptée lors de l'entrée dans la zone",
			['WHEN_KILL'] = "Acceptée en tuant :",
			['SEARCH_NEW'] = "Nouvelle",
			['SEARCH_CLEAR'] = "Effacer",
			['SEARCH_ALL_QUESTS'] = "Toutes les quêtes",
			['NEAR'] = "Proche",
			['FIRST_PREREQUISITE'] = "Première dans la suite de prérequis :",
			['BUGGED'] = "|cffff0000*** BOGUÉE ***|r",
			['IN_LOG'] = "Dans le journal",
			['TURNED_IN'] = "Rendue",
			['EVER_COMPLETED'] = "Has ever been completed",
			['ITEM'] = "Objet",
			['ITEM_LACK'] = "Objet manquant",
			['ABANDONED'] = "Abandonnée",
			['NEVER_ABANDONED'] = "Jamais abandonnée",
			['ACCEPTED'] = "Acceptée",
			['LEGENDARY'] = "Légendaire",
			}
    elseif "itIT" == locale then
		C[1][0]="Base Quests"
		C[2][0]="|cFF00FF00Completed|r"
		C[3][0]="|cFFFF0000Needs prerequisites|r"
		C[4][0]="|cFF996600Unobtainable|r"
		C[6][0]="|cFF0099CCRepeatable|r"
		C[8][0]="|cFFFF00FFIn Log|r"
		C[9][0]="|cFF666666Low-level|r"
		C[10][0]="High level"
		C[11][0]="Quest Title Appearance"
		C[12][0]="Prepend quest level"
		C[13][0]="Append required level"
		C[14][0]="Show whether repeatable quests previously completed"
		C[15][0]="Show status of quests in log"
		C[17][0]="Display map pins for quest givers"
		C[18][0]="Display button on world map"
		C[19][0]="Display dungeon quests in outer map"
		C[20][0]="Open world map updates when zones change"
		C[21][0]="Other"
		C[22][0]="Quest log panel updates when zones change"
		C[23][0]="Display breadcrumb quest information on Quest Frame"
		C[24][0]="Display holiday quests always"
		C[25][0]="Show only Loremaster quests"
		C[26][0]="Enable player coordinates"
		C[27][0]="Show achievement completion colors"
		C[28][0]="Bugged quests considered unobtainable"
		C[29][0]="Wide Wholly Quest Panel"
		C[30][0]="Show"
		C[31][0]='Show quest counts'
		C[32][0]='Live quest count updates'
		C[33][0]="Load Data"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['IS_BREADCRUMB'] = "Is breadcrumb quest for:",
			['PREREQUISITES'] = "Prerequisite quests:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "World Events",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Grail does not have this quest|r",
			['QUEST_ID'] = "Quest ID: ",
			['REQUIRED_LEVEL'] = "Required Level",
			['MAXIMUM_LEVEL_NONE'] = "None",
			['QUEST_TYPE_NORMAL'] = "Normal",
			['MAPAREA_NONE'] = "None",
			['LOREMASTER_AREA'] = "Loremaster Area",
			['FACTION_BOTH'] = "Both",
			['CLASS_NONE'] = "None",
			['CLASS_ANY'] = "Any",
			['GENDER_NONE'] = "None",
			['GENDER_BOTH'] = "Both",
			['GENDER'] = "Gender",
			['RACE_NONE'] = "None",
			['RACE_ANY'] = "Any",
			['HOLIDAYS_ONLY'] = "Available only during Holidays:",
			['SP_MESSAGE'] = "Special quest never enters Blizzard quest log",
			['INVALIDATE'] = "Invalidated by Quests:",
			['OAC'] = "On acceptance complete quests:",
			['OCC'] = "On completion of requirements complete quests:",
			['OTC'] = "On turn in complete quests:",
			['ENTER_ZONE'] = "Accepted when entering map area",
			['WHEN_KILL'] = "Accepted when killing:",
			['SEARCH_NEW'] = "New",
			['SEARCH_CLEAR'] = "Clear",
			['SEARCH_ALL_QUESTS'] = "All quests",
			['NEAR'] = "Near",
			['FIRST_PREREQUISITE'] = "First in Prerequisite Chain:",
			['BUGGED'] = "|cffff0000*** BUGGED ***|r",
			['IN_LOG'] = "In Log",
			['TURNED_IN'] = "Turned in",
			['EVER_COMPLETED'] = "Has ever been completed",
			['ITEM'] = "Item",
			['ITEM_LACK'] = "Item lack",
			['ABANDONED'] = "Abandoned",
			['NEVER_ABANDONED'] = "Never Abandoned",
			['ACCEPTED'] = "Accepted",
			['LEGENDARY'] = "Legendary",
			}
	elseif "koKR" == locale then
		C[1][0]="Base Quests"
		C[2][0]="|cFF00FF00Completed|r"
		C[3][0]="|cFFFF0000Needs prerequisites|r"
		C[4][0]="|cFF996600Unobtainable|r"
		C[6][0]="|cFF0099CC반복|r"
		C[8][0]="|cFFFF00FFIn Log|r"
		C[9][0]="|cFF666666Low-level|r"
		C[10][0]="High level"
		C[11][0]="Quest Title Appearance"
		C[12][0]="Prepend quest level"
		C[13][0]="Append required level"
		C[14][0]="Show whether repeatable quests previously completed"
		C[15][0]="Show status of quests in log"
		C[17][0]="Display map pins for quest givers"
		C[18][0]="세계지도에 버튼 표시"
		C[19][0]="Display dungeon quests in outer map"
		C[20][0]="Open world map updates when zones change"
		C[21][0]="Other"
		C[22][0]="Quest log panel updates when zones change"
		C[23][0]="Display breadcrumb quest information on Quest Frame"
		C[24][0]="Display holiday quests always"
		C[25][0]="Show only Loremaster quests"
		C[26][0]="Enable player coordinates"
		C[27][0]="Show achievement completion colors"
		C[28][0]="Bugged quests considered unobtainable"
		C[29][0]="Wide Wholly Quest Panel"
		C[30][0]="Show"
		C[31][0]='Show quest counts'
		C[32][0]='Live quest count updates'
		C[33][0]="Load Data"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['IS_BREADCRUMB'] = "Is breadcrumb quest for:",
			['PREREQUISITES'] = "Prerequisites:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "이벤트",
			['REPUTATION_REQUIRED'] = "평판 요구 사항:",
			['REPEATABLE'] = "반복",
			['YEARLY'] = "연간",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Grail does not have this quest|r",
			['QUEST_ID'] = "퀘스트 ID: ",
			['REQUIRED_LEVEL'] = "요구 레벨",
			['MAXIMUM_LEVEL_NONE'] = "None",
			['QUEST_TYPE_NORMAL'] = "Normal",
			['MAPAREA_NONE'] = "None",
			['LOREMASTER_AREA'] = "Loremaster Area",
			['FACTION_BOTH'] = "Both",
			['CLASS_NONE'] = "None",
			['CLASS_ANY'] = "Any",
			['GENDER_NONE'] = "None",
			['GENDER_BOTH'] = "Both",
			['GENDER'] = "Gender",
			['RACE_NONE'] = "None",
			['RACE_ANY'] = "Any",
			['HOLIDAYS_ONLY'] = "Available only during Holidays:",
			['SP_MESSAGE'] = "Special quest never enters Blizzard quest log",
			['INVALIDATE'] = "Invalidated by Quests:",
			['OAC'] = "On acceptance complete quests:",
			['OCC'] = "On completion of requirements complete quests:",
			['OTC'] = "On turn in complete quests:",
			['ENTER_ZONE'] = "Accepted when entering map area",
			['WHEN_KILL'] = "Accepted when killing:",
			['SEARCH_NEW'] = "New",
			['SEARCH_CLEAR'] = "Clear",
			['SEARCH_ALL_QUESTS'] = "All quests",
			['NEAR'] = "Near",
			['FIRST_PREREQUISITE'] = "First in Prerequisite Chain:",
			['BUGGED'] = "|cffff0000*** BUGGED ***|r",
			['IN_LOG'] = "In Log",
			['TURNED_IN'] = "Turned in",
			['EVER_COMPLETED'] = "Has ever been completed",
			['ITEM'] = "Item",
			['ITEM_LACK'] = "Item lack",
			['ABANDONED'] = "Abandoned",
			['NEVER_ABANDONED'] = "Never Abandoned",
			['ACCEPTED'] = "Accepted",
			['LEGENDARY'] = "Legendary",
			}
	elseif "ptBR" == locale then
		C[1][0]="Base Quests"
		C[2][0]="|cFF00FF00Completadas|r"
		C[3][0]="|cFFFF0000Pré-requisitos necessários|r"
		C[4][0]="|cFF996600Indisponível|r"
		C[6][0]="|cFF0099CCRepetíveis|r"
		C[8][0]="|cFFFF00FFNo registro|r"
		C[9][0]="|cFF666666Nível baixo|r"
		C[10][0]="Nível alto"
		C[11][0]="Quest Title Appearance"
		C[12][0]="Prefixar nível das missões"
		C[13][0]="Acrescentar nível requerido"
		C[14][0]="Mostrar quaisquer missões repetíveis já completadas"
		C[15][0]="Exibir status das missões no registro"
		C[17][0]="Marcar recrutadores no mapa"
		C[18][0]="Exibir botão no mapa mundi"
		C[19][0]="Exibir missões de masmorras no mapa exterior"
		C[20][0]="Abrir atualizações do mapa mundi quando trocar de zona"
		C[21][0]="Outros"
		C[22][0]="Painel de registro das missões atualiza quando mudar de zona"
		C[23][0]="Mostrar informação sobre missões guia no quadro de missões"
		C[24][0]="Sempre exibir missões de eventos"
		C[25][0]="Exibir somente missões do Mestre Historiador"
		C[26][0]="Habilitar coordenadas do jogador"
		C[27][0]="Show achievement completion colors"
		C[28][0]="Missões com erros consideradas indisponíveis"
		C[29][0]="Wide Wholly Quest Panel"
		C[30][0]="Show"
		C[31][0]='Show quest counts'
		C[32][0]='Live quest count updates'
		C[33][0]="Carregar dados"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Matar para começar [%s]",
			['DROP_TO_START_FORMAT'] = "Saqueie %s para começar [%s]",
			['REQUIRES_FORMAT'] = "Wholly requer a verão %s do Grail ou maior",
			['MUST_KILL_PIN_FORMAT'] = "%s [Matar]",
			['ESCORT'] = "Escolta",
			['BREADCRUMB'] = "Missão guia:",
			['IS_BREADCRUMB'] = "Is breadcrumb quest for:",
			['PREREQUISITES'] = "Missões requeridas:",
			['OTHER'] = "Outro",
			['SINGLE_BREADCRUMB_FORMAT'] = "Missão guia disponível",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Missão guia disponível",
			['WORLD_EVENTS'] = "Eventos Mundiais",
			['REPUTATION_REQUIRED'] = "Requer reputação:",
			['REPEATABLE'] = "Repetitível",
			['YEARLY'] = "Anual",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Grail não tem essa missão|r",
			['QUEST_ID'] = "ID da missão: ",
			['REQUIRED_LEVEL'] = "Requer nível",
			['MAXIMUM_LEVEL_NONE'] = "Nenhum",
			['QUEST_TYPE_NORMAL'] = "Normal",
			['MAPAREA_NONE'] = "Nenhuma",
			['LOREMASTER_AREA'] = "Área do Mestre Historiador",
			['FACTION_BOTH'] = "Ambas",
			['CLASS_NONE'] = "Nenhuma",
			['CLASS_ANY'] = "Qualquer",
			['GENDER_NONE'] = "Nenhum",
			['GENDER_BOTH'] = "Ambos",
			['GENDER'] = "Gênero",
			['RACE_NONE'] = "Nenhuma",
			['RACE_ANY'] = "Qualquer",
			['HOLIDAYS_ONLY'] = "Disponível apenas durante Eventos Mundiais:",
			['SP_MESSAGE'] = "Missões especiais nunca entram no registro da Blizzard",
			['INVALIDATE'] = "Invalidado pelas missões:",
			['OAC'] = "Após aceitar complete as missões:",
			['OCC'] = "Após completar os requisitos, complete as missões:",
			['OTC'] = "Quando entregar missões completas:",
			['ENTER_ZONE'] = "Aceita quando entra na área do mapa",
			['WHEN_KILL'] = "Aceita quando matar:",
			['SEARCH_NEW'] = "Nova",
			['SEARCH_CLEAR'] = "Limpar",
			['SEARCH_ALL_QUESTS'] = "Todas as missões",
			['NEAR'] = "Próximo",
			['FIRST_PREREQUISITE'] = "First in Prerequisite Chain:",
			['BUGGED'] = "|cffff0000*** COM ERRO ***|r",
			['IN_LOG'] = "In Log",
			['TURNED_IN'] = "Turned in",
			['EVER_COMPLETED'] = "Já foi concluída",
			['ITEM'] = "Item",
			['ITEM_LACK'] = "Falta de item",
			['ABANDONED'] = "Abandonada",
			['NEVER_ABANDONED'] = "Nunca abandonada",
			['ACCEPTED'] = "Aceita",
			['LEGENDARY'] = "Legendary",
			}
	elseif "ruRU" == locale then
		C[1][0]="База заданий"
		C[2][0]="|cFF00FF00Завершенные|r"
		C[3][0]="|cFFFF0000Требуются предварительные|r"
		C[4][0]="|cFF996600Недоступные|r"
		C[6][0]="|cFF0099CCПовторяющиеся|r"
		C[8][0]="|cFFFF00FFУже в журнале заданий|r"
		C[9][0]="|cFF666666Низкого уровня|r"
		C[10][0]="Высокого уровня"
		C[11][0]="Название задания"
		C[12][0]="Показывать уровень задания"
		C[13][0]="Показывать требуемый уровень"
		C[14][0]="Показывать ранее выполненные повторяемые задания"
		C[15][0]="Отображать в журнале статус заданий"
		C[17][0]="Показывать на карте мира метки тех, кто дает задания"
		C[18][0]="Отображать кнопку на карте мира"
		C[19][0]="Показывать задания в подземельях на карте игровой зоны"
		C[20][0]="Обновлять карту мира при смене игровой зоны"
		C[21][0]="Прочие"
		C[22][0]="Обновлять журнал заданий при смене игровой зоны"
		C[23][0]="Показывать информацию о наличии путеводных заданий"
		C[24][0]="Всегда показывать праздничные задания"
		C[25][0]="Показывать лишь задания, необходимые для достижения Хранителя мудрости"
		C[26][0]="Отображать координаты игрока"
		C[27][0]="Показывать завершенные достижения цветом"
		C[28][0]="Считать ошибочные задания невозможными для получения"
		C[29][0]="Wide Wholly Quest Panel"
		C[30][0]="Show"
		C[31][0]='Show quest counts'
		C[32][0]='Live quest count updates'
		C[33][0]="Загрузка данных"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Убить, чтобы начать [%s]",
			['DROP_TO_START_FORMAT'] = "Падает %s, начинает [%s]",
			['REQUIRES_FORMAT'] = "Для работы Wholly требуется Grail версии %s или выше",
			['MUST_KILL_PIN_FORMAT'] = "%s [Убить]",
			['ESCORT'] = "Сопровождение",
			['BREADCRUMB'] = "Путеводные задания:",
			['IS_BREADCRUMB'] = "Путеводное задание для:",
			['PREREQUISITES'] = "Предварительные задания:",
			['OTHER'] = "Другое",
			['SINGLE_BREADCRUMB_FORMAT'] = "Доступно путеводное задание",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "Доступно %d путеводных заданий",
			['WORLD_EVENTS'] = "Игровые события",
			['REPUTATION_REQUIRED'] = "Требуемая репутация:",
			['REPEATABLE'] = "Повторяющиеся",
			['YEARLY'] = "Ежегодные",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Этого задания нет в Grail|r",
			['QUEST_ID'] = "ID задания: ",
			['REQUIRED_LEVEL'] = "Требуемый уровень",
			['MAXIMUM_LEVEL_NONE'] = "Нет",
			['QUEST_TYPE_NORMAL'] = "Обычный",
			['MAPAREA_NONE'] = "Нет",
			['LOREMASTER_AREA'] = "Хранитель мудрости",
			['FACTION_BOTH'] = "Обе",
			['CLASS_NONE'] = "Нет",
			['CLASS_ANY'] = "Любой",
			['GENDER_NONE'] = "Нет",
			['GENDER_BOTH'] = "Оба",
			['GENDER'] = "Пол",
			['RACE_NONE'] = "Нет",
			['RACE_ANY'] = "Любая",
			['HOLIDAYS_ONLY'] = "Доступны только в праздничные дни:",
			['SP_MESSAGE'] = "Особый квест никогда не попадает в список заданий Blizzard",
			['INVALIDATE'] = "Недействительное из-за:",
			['OAC'] = "Задания, завершаемые при принятии:",
			['OCC'] = "Задания, завершаемые при выполнении условий:",
			['OTC'] = "Задания, завершаемые при возвращении:",
			['ENTER_ZONE'] = "Принимаемое при входе в игровую зону",
			['WHEN_KILL'] = "Принимаемые при убийстве:",
			['SEARCH_NEW'] = "Новый",
			['SEARCH_CLEAR'] = "Очистить",
			['SEARCH_ALL_QUESTS'] = "Все задания",
			['NEAR'] = "Рядом",
			['FIRST_PREREQUISITE'] = "Первое в цепочке предварительных:",
			['BUGGED'] = "|cffff0000*** ОШИБОЧНОЕ ***|r",
			['IN_LOG'] = "Уже в журнале заданий",
			['TURNED_IN'] = "Условия выполнены",
			['EVER_COMPLETED'] = "Был выполнен",
			['ITEM'] = "Предмет",
			['ITEM_LACK'] = "Предмет отсутствует",
			['ABANDONED'] = "Проваленный",
			['NEVER_ABANDONED'] = "Не отменялся",
			['ACCEPTED'] = "Принят",
			['LEGENDARY'] = "Legendary",
			}
	elseif "zhCN" == locale then
		C[1][0]="Base Quests"
		C[2][0]="|cFF00FF00Completed|r"
		C[3][0]="|cFFFF0000Needs prerequisites|r"
		C[4][0]="|cFF996600Unobtainable|r"
		C[6][0]="|cFF0099CCRepeatable|r"
		C[8][0]="|cFFFF00FFIn Log|r"
		C[9][0]="|cFF666666Low-level|r"
		C[10][0]="High level"
		C[11][0]="Quest Title Appearance"
		C[12][0]="Prepend quest level"
		C[13][0]="Append required level"
		C[14][0]="Show whether repeatable quests previously completed"
		C[15][0]="Show status of quests in log"
		C[17][0]="Display map pins for quest givers"
		C[18][0]="Display button on world map"
		C[19][0]="Display dungeon quests in outer map"
		C[20][0]="Open world map updates when zones change"
		C[21][0]="Other"
		C[22][0]="Quest log panel updates when zones change"
		C[23][0]="Display breadcrumb quest information on Quest Frame"
		C[24][0]="Display holiday quests always"
		C[25][0]="Show only Loremaster quests"
		C[26][0]="Enable player coordinates"
		C[27][0]="Show achievement completion colors"
		C[28][0]="Bugged quests considered unobtainable"
		C[29][0]="Wide Wholly Quest Panel"
		C[30][0]="Show"
		C[31][0]='Show quest counts'
		C[32][0]='Live quest count updates'
		C[33][0]="Load Data"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['IS_BREADCRUMB'] = "Is breadcrumb quest for:",
			['PREREQUISITES'] = "Prerequisites:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "World Events",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Grail does not have this quest|r",
			['QUEST_ID'] = "Quest ID: ",
			['REQUIRED_LEVEL'] = "Required Level",
			['MAXIMUM_LEVEL_NONE'] = "None",
			['QUEST_TYPE_NORMAL'] = "Normal",
			['MAPAREA_NONE'] = "None",
			['LOREMASTER_AREA'] = "Loremaster Area",
			['FACTION_BOTH'] = "Both",
			['CLASS_NONE'] = "None",
			['CLASS_ANY'] = "Any",
			['GENDER_NONE'] = "None",
			['GENDER_BOTH'] = "Both",
			['GENDER'] = "Gender",
			['RACE_NONE'] = "None",
			['RACE_ANY'] = "Any",
			['HOLIDAYS_ONLY'] = "Available only during Holidays:",
			['SP_MESSAGE'] = "Special quest never enters Blizzard quest log",
			['INVALIDATE'] = "Invalidated by Quests:",
			['OAC'] = "On acceptance complete quests:",
			['OCC'] = "On completion of requirements complete quests:",
			['OTC'] = "On turn in complete quests:",
			['ENTER_ZONE'] = "Accepted when entering map area",
			['WHEN_KILL'] = "Accepted when killing:",
			['SEARCH_NEW'] = "New",
			['SEARCH_CLEAR'] = "Clear",
			['SEARCH_ALL_QUESTS'] = "All quests",
			['NEAR'] = "Near",
			['FIRST_PREREQUISITE'] = "First in Prerequisite Chain:",
			['BUGGED'] = "|cffff0000*** BUGGED ***|r",
			['IN_LOG'] = "In Log",
			['TURNED_IN'] = "Turned in",
			['EVER_COMPLETED'] = "Has ever been completed",
			['ITEM'] = "Item",
			['ITEM_LACK'] = "Item lack",
			['ABANDONED'] = "Abandoned",
			['NEVER_ABANDONED'] = "Never Abandoned",
			['ACCEPTED'] = "Accepted",
			['LEGENDARY'] = "Legendary",
			}
	elseif "zhTW" == locale then
		C[1][0]="Base Quests"
		C[2][0]="|cFF00FF00Completed|r"
		C[3][0]="|cFFFF0000Needs prerequisites|r"
		C[4][0]="|cFF996600Unobtainable|r"
		C[6][0]="|cFF0099CCRepeatable|r"
		C[8][0]="|cFFFF00FFIn Log|r"
		C[9][0]="|cFF666666Low-level|r"
		C[10][0]="High level"
		C[11][0]="Quest Title Appearance"
		C[12][0]="Prepend quest level"
		C[13][0]="Append required level"
		C[14][0]="Show whether repeatable quests previously completed"
		C[15][0]="Show status of quests in log"
		C[17][0]="Display map pins for quest givers"
		C[18][0]="Display button on world map"
		C[19][0]="Display dungeon quests in outer map"
		C[20][0]="Open world map updates when zones change"
		C[21][0]="Other"
		C[22][0]="Quest log panel updates when zones change"
		C[23][0]="Display breadcrumb quest information on Quest Frame"
		C[24][0]="Display holiday quests always"
		C[25][0]="Show only Loremaster quests"
		C[26][0]="Enable player coordinates"
		C[27][0]="Show achievement completion colors"
		C[28][0]="Bugged quests considered unobtainable"
		C[29][0]="Wide Wholly Quest Panel"
		C[30][0]="Show"
		C[31][0]='Show quest counts'
		C[32][0]='Live quest count updates'
		C[33][0]="Load Data"
		Wholly.s = {
			['KILL_TO_START_FORMAT'] = "Kill to start [%s]",
			['DROP_TO_START_FORMAT'] = "Drops %s to start [%s]",
			['REQUIRES_FORMAT'] = "Wholly requires Grail version %s or later",
			['MUST_KILL_PIN_FORMAT'] = "%s [Kill]",
			['ESCORT'] = "Escort",
			['BREADCRUMB'] = "Breadcrumb quests:",
			['IS_BREADCRUMB'] = "Is breadcrumb quest for:",
			['PREREQUISITES'] = "Prerequisites:",
			['OTHER'] = "Other",
			['SINGLE_BREADCRUMB_FORMAT'] = "Breadcrumb quest available",
			['MULTIPLE_BREADCRUMB_FORMAT'] = "%d Breadcrumb quests available",
			['WORLD_EVENTS'] = "世界事件",
			['REPUTATION_REQUIRED'] = "Reputation Required:",
			['REPEATABLE'] = "Repeatable",
			['YEARLY'] = "Yearly",
			['GRAIL_NOT_HAVE'] = "|cFFFF0000Grail does not have this quest|r",
			['QUEST_ID'] = "Quest ID: ",
			['REQUIRED_LEVEL'] = "Required Level",
			['MAXIMUM_LEVEL_NONE'] = "None",
			['QUEST_TYPE_NORMAL'] = "Normal",
			['MAPAREA_NONE'] = "None",
			['LOREMASTER_AREA'] = "Loremaster Area",
			['FACTION_BOTH'] = "Both",
			['CLASS_NONE'] = "None",
			['CLASS_ANY'] = "Any",
			['GENDER_NONE'] = "None",
			['GENDER_BOTH'] = "Both",
			['GENDER'] = "Gender",
			['RACE_NONE'] = "None",
			['RACE_ANY'] = "Any",
			['HOLIDAYS_ONLY'] = "Available only during Holidays:",
			['SP_MESSAGE'] = "Special quest never enters Blizzard quest log",
			['INVALIDATE'] = "Invalidated by Quests:",
			['OAC'] = "On acceptance complete quests:",
			['OCC'] = "On completion of requirements complete quests:",
			['OTC'] = "On turn in complete quests:",
			['ENTER_ZONE'] = "Accepted when entering map area",
			['WHEN_KILL'] = "Accepted when killing:",
			['SEARCH_NEW'] = "New",
			['SEARCH_CLEAR'] = "Clear",
			['SEARCH_ALL_QUESTS'] = "All quests",
			['NEAR'] = "Near",
			['FIRST_PREREQUISITE'] = "First in Prerequisite Chain:",
			['BUGGED'] = "|cffff0000*** BUGGED ***|r",
			['IN_LOG'] = "In Log",
			['TURNED_IN'] = "Turned in",
			['EVER_COMPLETED'] = "Has ever been completed",
			['ITEM'] = "Item",
			['ITEM_LACK'] = "Item lack",
			['ABANDONED'] = "Abandoned",
			['NEVER_ABANDONED'] = "Never Abandoned",
			['ACCEPTED'] = "Accepted",
			['LEGENDARY'] = "Legendary",
			}
	end

end
